-- $Id: bwaa-aliasing.adb 186 2009-07-28 06:38:38Z georg $
---------------------------------------------------------------------------
--                                                                       --
--                   BWAA - Busy Wait Analyser for Ada                   --
--      BWAA is a program that finds busy waiting in Ada programs.       --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--                 Copyright (c) 2009 Georg Kienesberger                 --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of BWAA.                                            --
--                                                                       --
-- BWAA is free software; you can redistribute it and/or modify          --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- BWAA is distributed in the hope that it will be useful,               --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------
with Ada.Unchecked_Deallocation;

with Ast2Cfg.Flow_Types; use Ast2Cfg.Flow_Types;
with Ast2Cfg.Cfgs; use Ast2Cfg.Cfgs;
with Ast2Cfg.Pkgs; use Ast2Cfg.Pkgs;
with Ast2Cfg.Utils; use Ast2Cfg.Utils;
use type Ast2Cfg.Long_Long_Natural;

with Bwaa.Output; use Bwaa.Output;
with Bwaa.Visibility.Visibility_Info;

package body Bwaa.Aliasing is

   -- clear the contents of the alias info
   procedure Clear (Aliases: in out Alias_Info) is
   begin
      Clear(Aliases.Variable_Aliases);
      Clear(Aliases.Package_Aliases);
   end Clear;

   -- free allocated memory, set Alias_Info_Ptr null
   procedure Free (Aliases_Ptr: in out Alias_Info_Ptr) is
      procedure Destroy is new Ada.Unchecked_Deallocation(Alias_Info, Alias_Info_Ptr);
   begin
      Clear(Aliases_Ptr.Variable_Aliases);
      Clear(Aliases_Ptr.Package_Aliases);
      Destroy(Aliases_Ptr);
   end Free;

   -- completes a not fully qualified variable name e.g: Bar.X => Foo.Bar.X
   procedure Complete_Prefix(Var_Name: in out Unbounded_String;
                             Flow: in Ast2Cfg.Flow_Types.Flow_Object'Class;
                             Specs: in Bwaa.Util.Specification_Map.Map) is

      Head: Unbounded_String := Null_Unbounded_String;
      Succs: Flow_Class_Ptr_List.Object := Get_Succs(Flow);
      Preds: Flow_Class_Ptr_List.Object := Get_Preds(Flow);
      use Flow_Class_Ptr_List;
      Succ: Flow_Class_Ptr := null;
      Pred: Flow_Class_Ptr := null;
      Rens: Alias_Class_Ptr_List.Object := Get_Rens(Flow);
      Cur_Ren: Alias_Class_Ptr := null;
   begin

      -- warning: this does not respect visibility!
      if Count(Var_Name, ".") = 0 then
         -- Flow: Foo.Bar, Var: X => Foo.Bar.X
         Var_Name := Get_Full_Name(Flow) & "." & Var_Name;
         return;
      end if;

      Head := Ada.Strings.Unbounded.Head(Var_Name, Index(Var_Name, ".") - 1);

      if Head = Unbounded_String'(Get_Name(Flow)) then
         if Get_Prefix(Flow) /= Null_Unbounded_String then
            -- Flow: Foo.Bar, Var: Bar.X => Foo.Bar.X
            Var_Name := Get_Prefix(Flow) & "." & Var_Name;
         end if;
         -- Flow: Foo, Var: Foo.X => Foo.X
         return;
      end if;

      -- check package renamings
      Alias_Class_Ptr_List.Reset(Rens);
      while Alias_Class_Ptr_List.Has_Next(Rens) loop
         Alias_Class_Ptr_List.Get_Next(Rens, Cur_Ren);
         if Cur_Ren.all in Ast2Cfg.Flow_Types.Pkg_Alias'Class then
            if Head = Cur_Ren.New_Name then
               Var_Name := Unbounded_String'(Get_Name(Flow)) & "." & Var_Name;
               return;
            end if;
         end if;
      end loop;

      -- Flow: Foo.Bar, Succ: Sub, Var: Sub.X => Foo.Bar.Sub.X
      Reset(Succs);
      while Has_Next(Succs) loop
         Get_Next(Succs, Succ);

         if Head = Unbounded_String'(Get_Name(Succ.all)) then
            Var_Name := Get_Prefix(Succ.all) & "." & Var_Name;
            return;
         end if;
      end loop;

      -- Flow: Foo.Bar, Succ: Sub, Var: Sub.X => Foo.Bar.Sub.X, this time for spec
      declare
         Spec: Flow_Class_Ptr := null;
      begin
         Spec := Bwaa.Util.Specification_Map.Element(Specs, Get_Full_Name(Flow));
         Succs := Get_Succs(Spec.all);
         Reset(Succs);
         while Has_Next(Succs) loop
            Get_Next(Succs, Succ);

            if Head = Unbounded_String'(Get_Name(Succ.all)) then
               Var_Name := Get_Prefix(Succ.all) & "." & Var_Name;
               return;
            end if;
         end loop;
      exception
         when Constraint_Error =>
            Put_Line(Verbose, "[Bwaa.Aliasing] Unable to retrieve spec for " & U2s(Get_Full_Name(Flow)));
      end;

      -- recurse upwards
      if Count(Preds) = 0 then
         Put_Line(Warning, "[Bwaa.Aliasing] Warning: Unable to find complete prefix for " & U2s(Var_Name));
         return;
      elsif Count(Preds) = 1 then
         Pred := Get_Head(Preds);
         Complete_Prefix(Var_Name, Pred.all, Specs);
      else
         raise Invalid_Pkg_Cfg_Tree with "[Bwaa.Aliasing] ERROR: Flow with more than one predecessor encountered.";
      end if;

      Pred := Get_Head(Preds);

   end Complete_Prefix;

   -- collect all variables and their aliases from the specifications included by a with clause
   procedure Collect_From_With_Subtree (Aliases: in out Alias_Info;
                                        Flow: in Ast2Cfg.Flow_Types.Flow_Object'Class;
                                        Specs: in Bwaa.Util.Specification_Map.Map;
                                        Visi_Info: in Visibility_Info.Info) is

      Clauses: Ref_Class_Ptr_List.Object := Get_Refs(Flow);
      use Ref_Class_Ptr_List;
      Clause: Ref_Class_Ptr := null;
      Target: Flow_Class_Ptr := null;
   begin

      -- in case this is a subprogram within a default package the clauses are in the default package
      if Flow in Ast2Cfg.CFGs.CFG_Object'Class then
         declare
            Preds: Flow_Class_Ptr_List.Object := Get_Preds(Flow);
            Pred: Flow_Class_Ptr := null;
            use Flow_Class_Ptr_List;
         begin
            if Count(Preds) /= 1 then
               raise Invalid_Pkg_Cfg_Tree with "[Bwaa.Aliasing] ERROR: CFG_Object with more/less than one predecessor encountered.";
            end if;

            Pred := Get_Head(Preds);

            if Pred.all in Ast2Cfg.Pkgs.Def_Object'Class then
               Clauses := Get_Refs(Pred.all);
            end if;
         end;
      end if;

      Reset(Clauses);
      while Has_Next(Clauses) loop
         Get_Next(Clauses, Clause);
         if Clause.all in With_Ref'Class then
            declare
            begin
               Target := null;
               Target := Bwaa.Util.Specification_Map.Element(Specs, Merge_UStr_List(Clause.Name, "."));
            exception
               when Constraint_Error =>
                  Put_Line(Verbose, "[Bwaa.Aliasing] Unable to find target of with clause: " & U2s(Merge_UStr_List(Clause.Name, ".")));
            end;

            if Target /= null then
               if Target.all in Ast2Cfg.Pkgs.Spec_Object'Class then
                  Put_Line(Verbose, "[Bwaa.Aliasing] Adding aliases of package " & U2s(Get_Full_Name(Target.all)) & " (referenced by with clause).");
                  Collect_From_Spec_Subtree(Aliases, Ast2Cfg.Pkgs.Spec_Object(Target.all), Specs, Get_Full_Name(Target.all), Visi_Info);
               else
                  raise Wrong_Flow_Object_Exception with "[Bwaa.Aliasing] ERROR: Flow referenced by with clause is not a package spec.";
               end if;
            end if;

         end if;
      end loop;

   end Collect_From_With_Subtree;

   -- collect all variables and their aliases from the specifications in the subtree
   procedure Collect_From_Spec_Subtree (Aliases: in out Alias_Info;
                                        Spec: in Ast2Cfg.Pkgs.Spec_Object'Class;
                                        Specs: in Bwaa.Util.Specification_Map.Map;
                                        Root_Name: in Unbounded_String;
                                        Visi_Info: in Visibility_Info.Info) is

      Succs: Flow_Class_Ptr_List.Object := Get_Succs(Spec);
      Decls: Declaration_Ptr_List.Object := Get_Decls(Spec);
      Rens: Alias_Class_Ptr_List.Object := Get_Rens(Spec);
      Cur_Succ: Flow_Class_Ptr := null;
      Cur_Decl: Declaration_Ptr := null;
      Cur_Ren: Alias_Class_Ptr := null;
      Full_Spec_Name: UStr_List.Object;
   begin

      Get_Full_Name(Spec, Full_Spec_Name);

      Put_Line(Debug, "[Bwaa.Aliasing] Collecting aliases from spec subtree. Currently at: " & U2s(Get_Full_Name(Spec)) & " (spec)");

      -- get the variable declarations of the current flow object
      Declaration_Ptr_List.Reset(Decls);
      while Declaration_Ptr_List.Has_Next(Decls) loop
         Declaration_Ptr_List.Get_Next(Decls, Cur_Decl);
         declare
            Var_Name: Unbounded_String := Null_Unbounded_String;
         begin
            UStr_List.Reset(Cur_Decl.Vars);
            while Ustr_List.Has_Next(Cur_Decl.Vars) loop
               Ustr_List.Get_Next(Cur_Decl.Vars, Var_Name);
               Add_Prefixed_Series(Aliases.Variable_Aliases, Aliases, Var_Name, Full_Spec_Name, Root_Name);
               Put_Line(Debug, "[Bwaa.Aliasing] Declaration found: " & U2s(Var_Name));
            end loop;
         end;
      end loop;

      -- get the package declarations of the current flow object
      Flow_Class_Ptr_List.Reset(Succs);
      while Flow_Class_Ptr_List.Has_Next(Succs) loop
         Flow_Class_Ptr_List.Get_Next(Succs, Cur_Succ);
         if Cur_Succ.all in Ast2Cfg.Pkgs.Pkg_Object'Class then
            Add_Prefixed_Series(Aliases.Package_Aliases, Aliases, Get_Name(Cur_Succ.all), Full_Spec_Name, Root_Name);
            Put_Line(Debug, "[Bwaa.Aliasing] Declaration found: " & Get_Name(Cur_Succ.all));
         end if;
      end loop;

      -- get the renamings of the current flow object
      Alias_Class_Ptr_List.Reset(Rens);
      while Alias_Class_Ptr_List.Has_Next(Rens) loop
         Alias_Class_Ptr_List.Get_Next(Rens, Cur_Ren);

         if Cur_Ren.all in Ast2Cfg.Flow_Types.Obj_Alias'Class then
            declare
               Full_Old_Var_Name: Unbounded_String := Merge_UStr_List(Cur_Ren.Old_Name, ".");
            begin

               if Count(Full_Old_Var_Name, ".") = 0 then
                  Visibility_Info.Get_Full_Visible_Name(Full_Old_Var_Name, Cur_Ren.Decl_Id, Spec, Specs, Visi_Info.Specs, Visi_Info.Bodies);
               else
                  Complete_Prefix(Full_Old_Var_Name, Spec, Specs);
               end if;

               Add_Variable_Alias(Aliases, Full_Old_Var_Name, String'(Merge_UStr_List(Full_Spec_Name, ".")) & "." & Cur_Ren.New_Name);
               Add_Prefixed_Series(Aliases.Variable_Aliases, Aliases, Cur_Ren.New_Name, Full_Spec_Name, Root_Name);
            end;
         elsif Cur_Ren.all in Ast2Cfg.Flow_Types.Pkg_Alias'Class then
            declare
               Full_Old_Pkg_Name: Unbounded_String := Merge_UStr_List(Cur_Ren.Old_Name, ".");
            begin

               if Count(Full_Old_Pkg_Name, ".") = 0 then
                  Visibility_Info.Get_Full_Visible_Name(Full_Old_Pkg_Name, Cur_Ren.Decl_Id, Spec, Specs, Visi_Info.Pkg_Specs, Visi_Info.Pkg_Bodies);
               else
                  Complete_Prefix(Full_Old_Pkg_Name, Spec, Specs);
               end if;

               Add_Package_Alias(Aliases, Full_Old_Pkg_Name, String'(Merge_UStr_List(Full_Spec_Name, ".")) & "." & Cur_Ren.New_Name);
               Update_Package_Aliases(Aliases, Full_Old_Pkg_Name, String'(Merge_UStr_List(Full_Spec_Name, ".")) & "." & Cur_Ren.New_Name);
               Update_Variable_Aliases(Aliases, Full_Old_Pkg_Name, String'(Merge_UStr_List(Full_Spec_Name, ".")) & "." & Cur_Ren.New_Name);
               Add_Prefixed_Series(Aliases.Package_Aliases, Aliases, Cur_Ren.New_Name, Full_Spec_Name, Root_Name);
            end;
         else
            Put_Line(Debug, "[Bwaa.Aliasing] Unidentified renaming: " & Merge_UStr_List(Cur_Ren.Old_Name, ".") & " => " & U2s(Cur_Ren.New_Name));
         end if;
      end loop;

      -- get the variable declarations/renamings of the packages included by a with clause
      Collect_From_With_Subtree(Aliases, Spec, Specs, Visi_Info);

      --  declare
--           A: Alias_Info_Ptr := new Alias_Info;
--        begin
--           A.all := Aliases;
--           Put_Line(Max, Debug_Image(A));
--        end;

      -- recursion
      declare
         Succs: Flow_Class_Ptr_List.Object := Get_Succs(Spec);
         use Flow_Class_Ptr_List;
         Succ: Flow_Class_Ptr := null;
      begin
         Reset(Succs);
         while Has_Next(Succs) loop
            Get_Next(Succs, Succ);
            if Succ.all in Ast2Cfg.Pkgs.Spec_Object'Class then
               Collect_From_Spec_Subtree(Aliases, Ast2Cfg.Pkgs.Spec_Object(Succ.all), Specs, Root_Name, Visi_Info);
            end if;
         end loop;
      end;

   end Collect_From_Spec_Subtree;

   -- collect all variables and their aliases from the given flow object
   procedure Collect_From_Flow (Aliases: in out Alias_Info;
                                Specs: in Bwaa.Util.Specification_Map.Map;
                                Visi_Info: in Visibility_Info.Info;
                                Flow: in Ast2Cfg.Flow_Types.Flow_Object'Class) is

      Succs: Flow_Class_Ptr_List.Object := Get_Succs(Flow);
      Decls: Declaration_Ptr_List.Object := Get_Decls(Flow);
      Rens: Alias_Class_Ptr_List.Object := Get_Rens(Flow);
      Cur_Succ: Flow_Class_Ptr := null;
      Cur_Decl: Declaration_Ptr := null;
      Cur_Ren: Alias_Class_Ptr := null;
      Just_Added: USet.Set;
      Pkg_Just_Added: USet.Set;
      Full_Flow_Name: Unbounded_String := Get_Full_Name(Flow);
      Full_Flow_Name_Parts: UStr_List.Object;
   begin
      Get_Full_Name(Flow, Full_Flow_Name_Parts);

      if Flow in Ast2Cfg.Cfgs.Spec_Object'Class or else
         Flow in Ast2Cfg.Pkgs.Spec_Object'Class then
         Put_Line(Debug, "[Bwaa.Aliasing] Collecting aliases from: " & U2s(Get_Full_Name(Flow)) & " (spec)");
      else
         Put_Line(Debug, "[Bwaa.Aliasing] Collecting aliases from: " & U2s(Get_Full_Name(Flow)) & " (body)");
      end if;

      if not (Flow in Ast2Cfg.Cfgs.Cfg_Object'Class) and then
         not (Flow in Ast2Cfg.Pkgs.Pkg_Object'Class) then
         raise Wrong_Flow_Object_Exception;
      end if;

      -- get the parameter declarations of the current flow object in case it is a subprogram
      if Flow in Ast2Cfg.CFGs.Body_Object'Class then
         declare
            Params: Declaration_Ptr_List.Object := Ast2Cfg.CFGs.Get_Params(Ast2Cfg.CFGs.CFG_Object(Flow));
            Cur_Param: Declaration_Ptr := null;
            Param_Name: Unbounded_String := Null_Unbounded_String;
         begin
            Declaration_Ptr_List.Reset(Params);
            while Declaration_Ptr_List.Has_Next(Params) loop
               Declaration_Ptr_List.Get_Next(Params, Cur_Param);
               UStr_List.Reset(Cur_Param.Vars);
               while Ustr_List.Has_Next(Cur_Param.Vars) loop
                  Ustr_List.Get_Next(Cur_Param.Vars, Param_Name);

                  Add_Prefixed_Series(Aliases.Variable_Aliases, Aliases, Param_Name, Full_Flow_Name_Parts);

                  if not Contains(Aliases.Variable_Aliases, Param_Name) then
                     Makeset(Aliases.Variable_Aliases, Param_Name);
                     Union(Aliases.Variable_Aliases, Param_Name, Get_Full_Name(Flow) & "." & Param_Name);
                  end if;

                  Put_Line(Debug, "[Bwaa.Aliasing] Parameter declaration found: " & U2s(Param_Name));
               end loop;
            end loop;
         end;
      end if;

      -- get the variable declarations of the current flow object
      Declaration_Ptr_List.Reset(Decls);
      while Declaration_Ptr_List.Has_Next(Decls) loop
         Declaration_Ptr_List.Get_Next(Decls, Cur_Decl);
         declare
            Var_Name: Unbounded_String := Null_Unbounded_String;
         begin
            UStr_List.Reset(Cur_Decl.Vars);
            while Ustr_List.Has_Next(Cur_Decl.Vars) loop
               Ustr_List.Get_Next(Cur_Decl.Vars, Var_Name);

               Add_Prefixed_Series(Aliases.Variable_Aliases, Aliases, Var_Name, Full_Flow_Name_Parts);

               if not Contains(Aliases.Variable_Aliases, Var_Name) then
                  Makeset(Aliases.Variable_Aliases, Var_Name);
                  Union(Aliases.Variable_Aliases, Var_Name, Get_Full_Name(Flow) & "." & Var_Name);
               end if;

               Put_Line(Debug, "[Bwaa.Aliasing] Declaration found: " & U2s(Var_Name));
            end loop;
         end;
      end loop;

      -- get the package declarations of the current flow object
      Flow_Class_Ptr_List.Reset(Succs);
      while Flow_Class_Ptr_List.Has_Next(Succs) loop
         Flow_Class_Ptr_List.Get_Next(Succs, Cur_Succ);
         if Cur_Succ.all in Ast2Cfg.Pkgs.Pkg_Object'Class then

            Add_Prefixed_Series(Aliases.Package_Aliases, Aliases, Get_Name(Cur_Succ.all), Full_Flow_Name_Parts);

            declare
               Pkg_Name: Unbounded_String := Get_Name(Cur_Succ.all);
            begin
               if not Contains(Aliases.Package_Aliases, Pkg_Name) then
                  Makeset(Aliases.Package_Aliases, Pkg_Name);
                  Union(Aliases.Package_Aliases, Pkg_Name, Get_Full_Name(Cur_Succ.all));
               end if;
            end;

            Put_Line(Debug, "[Bwaa.Aliasing] Declaration found: " & Get_Name(Cur_Succ.all));
         end if;
      end loop;

      -- get the variable declarations/renamings of the packages contained in this flow
      declare
         Succs: Flow_Class_Ptr_List.Object := Get_Succs(Flow);
         use Flow_Class_Ptr_List;
         Succ: Flow_Class_Ptr := null;
      begin
         Reset(Succs);
         while Has_Next(Succs) loop
            Get_Next(Succs, Succ);
            if Succ.all in Ast2Cfg.Pkgs.Spec_Object'Class then
               Collect_From_Spec_Subtree(Aliases, Ast2Cfg.Pkgs.Spec_Object(Succ.all), Specs, Get_Name(Succ.all), Visi_Info);
            end if;
         end loop;
      end;

      -- get the variable declarations/renamings of the packages included by a with clause
      Collect_From_With_Subtree(Aliases, Flow, Specs, Visi_Info);

      -- get the renamings of the current flow object
      Alias_Class_Ptr_List.Reset(Rens);
      while Alias_Class_Ptr_List.Has_Next(Rens) loop
         Alias_Class_Ptr_List.Get_Next(Rens, Cur_Ren);

         declare
            Full_Old_Var_Name: Unbounded_String := Merge_UStr_List(Cur_Ren.Old_Name, ".");
         begin

            if Cur_Ren.all in Ast2Cfg.Flow_Types.Obj_Alias'Class then

               if Count(Full_Old_Var_Name, ".") = 0 then
                  if Flow in Ast2Cfg.Cfgs.Spec_Object'Class or else
                     Flow in Ast2Cfg.Pkgs.Spec_Object'Class then

                     Visibility_Info.Get_Full_Visible_Name(Full_Old_Var_Name, Cur_Ren.Decl_Id, Flow, Specs, Visi_Info.Specs, Visi_Info.Bodies);

                  elsif Flow in Ast2Cfg.Cfgs.Body_Object'Class or else
                        Flow in Ast2Cfg.Pkgs.Body_Object'Class then

                     Visibility_Info.Get_Full_Visible_Name(Full_Old_Var_Name, Cur_Ren.Decl_Id, Flow, Specs, Visi_Info.Bodies, Visi_Info.Bodies);
                  else
                     raise Wrong_Flow_Object_Exception with "[Bwaa.Aliasing] ERROR: Flow is neither spec nor body.";
                  end if;
               else
                  Complete_Prefix(Full_Old_Var_Name, Flow, Specs);
               end if;

               if not Contains(Aliases.Variable_Aliases, Cur_Ren.New_Name) then

                  if (not Contains(Aliases.Variable_Aliases, Merge_UStr_List(Cur_Ren.Old_Name, ".")) or else
                     USet.Contains(Just_Added, Merge_UStr_List(Cur_Ren.Old_Name, "."))) then

                     Add_Variable_Alias(Aliases, Merge_UStr_List(Cur_Ren.Old_Name, "."), Cur_Ren.New_Name);
                     USet.Include(Just_Added, Cur_Ren.New_Name);
                  end if;

                  Add_Variable_Alias(Aliases, Full_Old_Var_Name, Cur_Ren.New_Name);
                  USet.Include(Just_Added, Cur_Ren.New_Name);

               end if;

               Add_Variable_Alias(Aliases, Full_Old_Var_Name, Full_Flow_Name & "." & Cur_Ren.New_Name);
               USet.Include(Just_Added, Full_Flow_Name & "." & Cur_Ren.New_Name);

               declare
                  Full_Flow_Name: UStr_List.Object;
               begin
                  Get_Full_Name(Flow, Full_Flow_Name);
                  Add_Prefixed_Series(Aliases.Variable_Aliases, Aliases, Cur_Ren.New_Name, Full_Flow_Name, Just_Added);
               end;

            elsif Cur_Ren.all in Ast2Cfg.Flow_Types.Pkg_Alias'Class then

               if Count(Full_Old_Var_Name, ".") = 0 then
                  if Flow in Ast2Cfg.Cfgs.Spec_Object'Class or else
                     Flow in Ast2Cfg.Pkgs.Spec_Object'Class then

                     Visibility_Info.Get_Full_Visible_Name(Full_Old_Var_Name, Cur_Ren.Decl_Id, Flow, Specs, Visi_Info.Pkg_Specs, Visi_Info.Pkg_Bodies);

                  elsif Flow in Ast2Cfg.Cfgs.Body_Object'Class or else
                    Flow in Ast2Cfg.Pkgs.Body_Object'Class then

                     Visibility_Info.Get_Full_Visible_Name(Full_Old_Var_Name, Cur_Ren.Decl_Id, Flow, Specs, Visi_Info.Pkg_Bodies, Visi_Info.Pkg_Bodies);
                  else
                     raise Wrong_Flow_Object_Exception with "[Bwaa.Aliasing] ERROR: Flow is neither spec nor body.";
                  end if;
               else
                  Complete_Prefix(Full_Old_Var_Name, Flow, Specs);
               end if;

               if not Contains(Aliases.Package_Aliases, Cur_Ren.New_Name) then

                  if (not Contains(Aliases.Package_Aliases, Merge_UStr_List(Cur_Ren.Old_Name, ".")) or else
                     USet.Contains(Pkg_Just_Added, Merge_UStr_List(Cur_Ren.Old_Name, "."))) then

                     Add_Package_Alias(Aliases, Merge_UStr_List(Cur_Ren.Old_Name, "."), Cur_Ren.New_Name);
                     Update_Package_Aliases(Aliases, Merge_UStr_List(Cur_Ren.Old_Name, "."), Cur_Ren.New_Name);
                     Update_Variable_Aliases(Aliases, Merge_UStr_List(Cur_Ren.Old_Name, "."), Cur_Ren.New_Name);
                     USet.Include(Pkg_Just_Added, Cur_Ren.New_Name);
                  end if;

                  Add_Package_Alias(Aliases, Full_Old_Var_Name, Cur_Ren.New_Name);
                  Update_Package_Aliases(Aliases, Full_Old_Var_Name, Cur_Ren.New_Name);
                  Update_Variable_Aliases(Aliases, Full_Old_Var_Name, Cur_Ren.New_Name);
                  USet.Include(Pkg_Just_Added, Cur_Ren.New_Name);

               end if;

               Add_Package_Alias(Aliases, Full_Old_Var_Name, Full_Flow_Name & "." & Cur_Ren.New_Name);
               USet.Include(Pkg_Just_Added, Full_Flow_Name & "." & Cur_Ren.New_Name);
               Update_Package_Aliases(Aliases, Full_Old_Var_Name, Full_Flow_Name & "." & Cur_Ren.New_Name);
               Update_Variable_Aliases(Aliases, Full_Old_Var_Name, Full_Flow_Name & "." & Cur_Ren.New_Name);
               declare
                  Full_Flow_Name: UStr_List.Object;
               begin
                  Get_Full_Name(Flow, Full_Flow_Name);
                  Add_Prefixed_Series(Aliases.Package_Aliases, Aliases, Cur_Ren.New_Name, Full_Flow_Name, Pkg_Just_Added);
               end;

            else
               Put_Line(Debug, "[Bwaa.Aliasing] Unidentified renaming found: " & Merge_UStr_List(Cur_Ren.Old_Name, ".") & " => " & U2s(Cur_Ren.New_Name));
            end if;
         end;
      end loop;

      --  declare
--           A: Alias_Info_Ptr := new Alias_Info;
--        begin
--           A.all := Aliases;
--           Put_Line(Max, Debug_Image(A));
--        end;

   end Collect_From_Flow;

   -- collect all variables and their aliases starting at the given flow object by going up the pkg/cfg tree
   procedure Collect (Aliases: in out Alias_Info;
                      Flow: in Ast2Cfg.Flow_Types.Flow_Object'Class;
                      Specs: in Bwaa.Util.Specification_Map.Map;
                      Visi_Info: in Visibility_Info.Info) is

      use Bwaa.Util.Specification_Map;
      Preds: Flow_Class_Ptr_List.Object := Get_Preds(Flow);
      Spec: Flow_Class_Ptr := null;
      Cur_Name: Unbounded_String := Get_Name(Flow);
   begin

      if Flow in Ast2Cfg.Pkgs.Def_Object'Class then
         return;
      end if;

      Collect_From_Flow(Aliases, Specs, Visi_Info, Flow);

      if Flow in Ast2Cfg.Pkgs.Body_Object'Class and
         Cur_Name /= Null_Unbounded_String and
         not (Flow in Ast2Cfg.Pkgs.Def_Object'Class) then

         declare
         begin
            Spec := Element(Specs, Get_Full_Name(Flow));
         exception
            when Constraint_Error =>
               Put_Line(Verbose, "[Bwaa.Aliasing] Found no spec for " & U2s(Get_Full_Name(Flow)));
         end;

         Collect_From_Flow(Aliases, Specs, Visi_Info, Spec.all);
      end if;

      if Flow_Class_Ptr_List.Count(Preds) = 0 then
         return;
      elsif Flow_Class_Ptr_List.Count(Preds) > 1 then
         raise Invalid_Pkg_Cfg_Tree with "[Bwaa.Aliasing] ERROR: Pkg/CFG with more than one predecessor encountered.";
      end if;

      Collect(Aliases, Flow_Class_Ptr_List.Get_Head(Preds).all, Specs, Visi_Info);

   end Collect;

   -- check whether both variable names represent the same variable
   function Equal_Variable_Names (Aliases_Ptr: in Alias_Info_Ptr; X: in Unbounded_String; Y: in Unbounded_String) return Boolean is
      Fx, Fy: Unbounded_String := Null_Unbounded_String;
      Va: Sets renames Aliases_Ptr.Variable_Aliases;
   begin

      declare
      begin
         Find(Va, X, Fx);
      exception
         when Not_In_Any_Set_Exception =>
            Put_Line(Warning, "[Bwaa.Aliasing] Warning: No aliasing information for variable " & U2s(X));
            return X = Y;
      end;

      declare
      begin
         Find(Va, Y, Fy);
      exception
         when Not_In_Any_Set_Exception =>
            Put_Line(Warning, "[Bwaa.Aliasing] Warning: No aliasing information for variable " & U2s(Y));
            return X = Y;
      end;

      return Fx = Fy;
   end Equal_Variable_Names;

   -- check whether both package names represent the same package
   procedure Equal_Package_Names (Aliases: in out Alias_Info; X: in Unbounded_String; Y: in Unbounded_String; Result: out Boolean) is
      Fx, Fy: Unbounded_String := Null_Unbounded_String;
      Pa: Sets renames Aliases.Package_Aliases;
   begin

      declare
      begin
         Find(Pa, X, Fx);
      exception
         when Not_In_Any_Set_Exception =>
            -- Put_Line(Max, "[Bwaa.Aliasing] No aliasing information for package " & U2s(X));
            Result := X = Y;
            return;
      end;

      declare
      begin
         Find(Pa, Y, Fy);
      exception
         when Not_In_Any_Set_Exception =>
            -- Put_Line(Max, "[Bwaa.Aliasing] No aliasing information for package " & U2s(Y));
            Result := X = Y;
            return;
      end;

      Result := Fx = Fy;
   end Equal_Package_Names;

   procedure Circular_Package_Names (Aliases: in out Alias_Info; Old_Name: in Unbounded_String; New_Name: in Unbounded_String; Result: out Boolean) is
      use Bwaa.Util;
      use Bwaa.Util.Unbounded_String_Set;
      Partition: Unbounded_String_Array := Split(New_Name);
      Cur_Stop: Positive := Partition'First + 1;
      Cur_Name: Unbounded_String := Null_Unbounded_String;
      Equal: Boolean := False;
      MAX_CIRCLES: constant Natural := 1;
      Circles: Natural := 0;
      To_Check: Unbounded_String_Set.Set := Unbounded_String_Set.Empty_Set;
      Cur_Check: Unbounded_String_Set.Cursor := Unbounded_String_Set.No_Element;
   begin

      --Put_Line(Max, "Is circular?: " & Bwaa.Util.U2s(Old_Name) & " => " & Bwaa.Util.U2s(New_Name));

      -- check for circles in the new name itself based on the existing aliases
      Insert(To_Check, Partition(Partition'First));
      while Cur_Stop <= Partition'Last loop
         Cur_Name := Merge(Partition, Cur_Stop);

         Cur_Check := First(To_Check);
         while Cur_Check /= Unbounded_String_Set.No_Element loop
            Equal_Package_Names(Aliases, Element(Cur_Check), Cur_Name, Equal);
            if Equal then
               --Put_Line(Max, "Circle: " & Bwaa.Util.U2s(Element(Cur_Check)) & ", " & Bwaa.Util.U2s(Cur_Name));
               Circles := Circles + 1;
               if Circles > MAX_CIRCLES then
                  Result := True;
                  return;
               end if;
            end if;
            Next(Cur_Check);
         end loop;
         Include(To_Check, Cur_Name);
         Cur_Stop := Cur_Stop + 1;
      end loop;

      -- heuristics based on the old name
      if Index(New_Name, Bwaa.Util.U2s(Old_Name), 1) = 1 and
         (Length(New_Name) <= Length(Old_Name) or else
          Element(New_Name, Length(Old_Name) + 1) = '.') then
         declare
            Tail: Unbounded_String := Merge(Partition, Partition'Last, Partition'First + Count(Old_Name, ".") + 1);
         begin
            if Index(Old_Name, Bwaa.Util.U2s(Tail), 1) + Length(Tail) - 1 = Length(Old_Name) then
               Result := True;
               return;
            end if;
         end;
      end if;

      Result := False;
   end Circular_Package_Names;

   -- debug image of aliases
   function Debug_Image (Aliases: in Alias_Info_Ptr) return String is
      Image: Unbounded_String := Null_Unbounded_String;
      Cur: Alias_Sets.Cursor;
      package USet renames Bwaa.Util.Unbounded_String_Set;

      procedure Sets_Debug_Image (Aliases: in out Alias_Sets.Sets; Image: in out Unbounded_String) is
         Canonicals: USet.Set;
         Canonical: Unbounded_String := Null_Unbounded_String;
         Cur_Canonical: USet.Cursor;
         use type USet.Cursor;
      begin

         -- collect all canonical elements
         -- * note that due to path compression every parent is the canonical element after this loop
         Cur := First(Aliases);
         while Cur /= No_Element loop
            Find(Aliases, Element(Cur), Canonical);
            USet.Include(Canonicals, Canonical);
            Next(Cur);
         end loop;

         Cur_Canonical := USet.First(Canonicals);
         while Cur_Canonical /= USet.No_Element loop
            Canonical := USet.Element(Cur_Canonical);
            Cur := First(Aliases);
            while Cur /= No_Element loop
               if Parent(Cur) = Canonical then -- see above comment *
                  Image := Image & ASCII.HT & Element(Cur) & ASCII.LF;
               end if;
               Next(Cur);
            end loop;
            Image := Image & ASCII.LF;
            USet.Next(Cur_Canonical);
         end loop;

      end Sets_Debug_Image;

   begin
      Image := Image & "Package Aliases: " & ASCII.LF & ASCII.LF;
      Sets_Debug_Image(Aliases.Package_Aliases, Image);

      Image := Image & "Variable Aliases: " & ASCII.LF & ASCII.LF;
      Sets_Debug_Image(Aliases.Variable_Aliases, Image);

      return U2s(Image);
   end Debug_Image;

   procedure Add_Prefixed_Series (Alias_Set: in out Alias_Sets.Sets;
                                  Aliases: in out Alias_Info;
                                  X: in Unbounded_String;
                                  Prefix_List: in out Ast2Cfg.Utils.UStr_List.Object;
                                  Just_Added: in out USet.Set;
                                  Lowest_Prefix: in Unbounded_String := Null_Unbounded_String) is

      Var_Name_Combination: Unbounded_String := X;
      Prefix_Combination: Unbounded_String := Null_Unbounded_String;
      Name_Slice: Unbounded_String := Null_Unbounded_String;
      Full_Name: Unbounded_String := Unbounded_String'(Merge_UStr_List(Prefix_List, ".")) & "." &X;
      Add: Boolean := False;
      Prefixes: UStr_List.Object;
   begin

      --  declare
--           A: Alias_Info_Ptr := new Alias_Info;
--        begin
--           A.all := Aliases;
--           Put_Line(Max, Debug_Image(A));
--        end;
      --Put_Line(Max, "$$$$$$$$$$$$$$ " & Bwaa.Util.U2s(X) & ", " & Bwaa.Util.U2s(Merge_UStr_List(Prefix_List, ".")) & ", " & Bwaa.Util.U2s(Lowest_Prefix));


      -- in any case insert the full name
      if not Contains(Alias_Set, Full_Name) then
         Makeset(Alias_Set, Full_Name);
         USet.Include(Just_Added, Full_Name);
      end if;

      UStr_List.Duplicate(Prefix_List, Prefixes);
      UStr_List.Reset(Prefixes);

      if Lowest_Prefix = Null_Unbounded_String then
         Add := True;
      end if;

      while not Ustr_List.Empty(Prefixes) loop
         UStr_List.Pop(Prefixes, Name_Slice);
         exit when Name_Slice = Null_Unbounded_String;

         Var_Name_Combination := Name_Slice & "." & Var_Name_Combination;

         if Prefix_Combination /= Null_Unbounded_String then
            Prefix_Combination := Name_Slice & "." & Prefix_Combination;
         else
            Prefix_Combination := Name_Slice;
         end if;

         if Name_Slice = Lowest_Prefix then
            Add := True;
         end if;

         if Add then
            --Put_Line(Max, "############# " & Bwaa.Util.U2s(Var_Name_Combination));
            if not Contains(Alias_Set, Var_Name_Combination) then
               Makeset(Alias_Set, Var_Name_Combination);
               USet.Include(Just_Added, Var_Name_Combination);
               -- link every variable name combination to the full name
               Union(Alias_Set, Full_Name, Var_Name_Combination);

               -- if this is a package alias, we also have to update already added aliases
               if Alias_Set = Aliases.Package_Aliases then
                  Update_Package_Aliases(Aliases, Full_Name, Var_Name_Combination);
                  Update_Variable_Aliases(Aliases, Full_Name, Var_Name_Combination);
               end if;
            end if;

            -- also add the aliases induced by existing package aliases
            declare
               use Alias_Sets;
               Pa: Alias_Sets.Sets renames Aliases.Package_Aliases;
               Cur: Cursor := No_Element;
               Alias: Unbounded_String := Null_Unbounded_String;
               Aliased_Combination: Unbounded_String := Null_Unbounded_String;
               Circular: Boolean := False;
            begin
               --Put_Line(Max, "!!!!!!!!!!!!! " & Bwaa.Util.U2s(Prefix_Combination));
               First(Pa, Prefix_Combination, Cur);
               while Cur /= No_Element loop
                  Alias := Element(Cur);
                  Aliased_Combination := Alias & "." & X;

                  if Aliased_Combination /= Var_Name_Combination then

                     Circular_Package_Names(Aliases, Var_Name_Combination, Aliased_Combination, Circular);
                     if Circular then
                        Put_Line(Max, "[Bwaa.Aliasing] Adding prefixed series: Skipped circular variable aliases: " & Bwaa.Util.U2s(Var_Name_Combination) & " => " & Bwaa.Util.U2s(Aliased_Combination));
                     else
                        Put_Line(Max, "[Bwaa.Aliasing] Adding prefixed series: Adding alias " & Bwaa.Util.U2s(Var_Name_Combination) & " => " & Bwaa.Util.U2s(Aliased_Combination));

                        if not Contains(Alias_Set, Aliased_Combination) then
                           Makeset(Alias_Set, Aliased_Combination);
                           USet.Include(Just_Added, Aliased_Combination);
                        end if;

                        Union(Alias_Set, Aliased_Combination, Var_Name_Combination);
                        --Put_Line(Max, "%%%%%%%%%%%%% " & Bwaa.Util.U2s(Var_Name_Combination) & " => " & Bwaa.Util.U2s(Aliased_Combination));

                        -- if this is a package alias, we also have to update already added aliases
                        if Alias_Set = Aliases.Package_Aliases then
                           Update_Package_Aliases(Aliases, Var_Name_Combination, Aliased_Combination);
                           Update_Variable_Aliases(Aliases, Var_Name_Combination, Aliased_Combination);
                        end if;
                     end if;
                  end if;

                  Next(Pa, Cur);
               end loop;
            end;
         end if;

      end loop;
   end Add_Prefixed_Series;

   procedure Add_Prefixed_Series (Alias_Set: in out Alias_Sets.Sets;
                                  Aliases: in out Alias_Info;
                                  X: in Unbounded_String;
                                  Prefix_List: in out Ast2Cfg.Utils.UStr_List.Object;
                                  Lowest_Prefix: in Unbounded_String := Null_Unbounded_String) is
      Just_Added: USet.Set;
   begin
      Add_Prefixed_Series(Alias_Set, Aliases, X, Prefix_List, Just_Added, Lowest_Prefix);
   end Add_Prefixed_Series;

   procedure Add_Variable_Alias (Aliases: in out Alias_Info; Old_Name: in Unbounded_String; New_Name: in Unbounded_String) is
      Va: Alias_Sets.Sets renames Aliases.Variable_Aliases;
   begin
      if not Contains(Va, Old_Name) then
         Makeset(Va, Old_Name);
      end if;

      if not Contains(Va, New_Name) then
         Makeset(Va, New_Name);
      end if;

      Union(Va, Old_Name, New_Name);
      Put_Line(Debug, "[Bwaa.Aliasing] Added variable alias: " & U2s(Old_Name) & " => " & U2s(New_Name));
   end Add_Variable_Alias;

   procedure Add_Package_Alias (Aliases: in out Alias_Info; Old_Name: in Unbounded_String; New_Name: in Unbounded_String) is
      Pa: Alias_Sets.Sets renames Aliases.Package_Aliases;
      Va: Alias_Sets.Sets renames Aliases.Variable_Aliases;
   begin

      if not Contains(Pa, Old_Name) then
         Makeset(Pa, Old_Name);
      end if;

      if not Contains(Pa, New_Name) then
         Makeset(Pa, New_Name);
      end if;

      Union(Pa, Old_Name, New_Name);
      Put_Line(Debug, "[Bwaa.Aliasing] Added package alias: " &U2s(Old_Name) & " => " & U2s(New_Name));

   end Add_Package_Alias;

   procedure Update_Package_Aliases (Aliases: in out Alias_Info; Old_Pkg_Name: in Unbounded_String; New_Pkg_Name: in Unbounded_String) is
      use Alias_Sets;
      use Bwaa.Util;
      Pa: Alias_Sets.Sets renames Aliases.Package_Aliases;
      Cur: Cursor := First(Pa);
      Old_Name: Unbounded_String := Null_Unbounded_String;
      New_Name: Unbounded_String := Null_Unbounded_String;
      Circular: Boolean := False;
   begin

      Put_Line(Max, "[Bwaa.Aliasing] Update_Package_Aliases called with: " & Bwaa.Util.U2s(Old_Pkg_Name) & ", " & Bwaa.Util.U2s(New_Pkg_Name));

      while Cur /= No_Element loop
         Old_Name := Element(Cur);
         if Index(Old_Name, Bwaa.Util.U2s(Old_Pkg_Name), 1) = 1 and
            not (Index(Old_Name, Bwaa.Util.U2s(New_Pkg_Name), 1) = 1) and
            (Length(Old_Name) <= Length(Old_Pkg_Name) or else
             Element(Old_Name, Length(Old_Pkg_Name) + 1) = '.') then

            New_Name := Replace_Slice(Old_Name, 1, Length(Old_Pkg_Name), Bwaa.Util.U2s(New_Pkg_Name));
            Circular_Package_Names(Aliases, Old_Name, New_Name, Circular);
            if Circular then
               Put_Line(Max, "[Bwaa.Aliasing] Skipped circular package aliases: " & Bwaa.Util.U2s(Old_Name) & " => " & Bwaa.Util.U2s(New_Name));
            else
               Put_Line(Max, "[Bwaa.Aliasing] Updating package alias " & Bwaa.Util.U2s(Old_Name) & " with " & Bwaa.Util.U2s(Old_Pkg_Name) & " => " & Bwaa.Util.U2s(New_Pkg_Name));
               Add_Package_Alias(Aliases, Old_Name, New_Name);
            end if;
         end if;
         Next(Cur);
      end loop;
   end Update_Package_Aliases;


   procedure Update_Variable_Aliases (Aliases: in out Alias_Info; Old_Pkg_Name: in Unbounded_String; New_Pkg_Name: in Unbounded_String) is
      use Alias_Sets;
      use Bwaa.Util;
      Va: Alias_Sets.Sets renames Aliases.Variable_Aliases;
      Pa: Alias_Sets.Sets renames Aliases.Package_Aliases;
      Cur: Cursor := First(Va);
      Old_Name: Unbounded_String := Null_Unbounded_String;
      New_Name: Unbounded_String := Null_Unbounded_String;
      Old_Prefix: Unbounded_String := Null_Unbounded_String;
      New_Prefix: Unbounded_String := Null_Unbounded_String;
      Variable_Name: Unbounded_String := Null_Unbounded_String;
      Circular: Boolean := False;
   begin

      Put_Line(Max, "[Bwaa.Aliasing] Update_Variable_Aliases called with: " & Bwaa.Util.U2s(Old_Pkg_Name) & ", " & Bwaa.Util.U2s(New_Pkg_Name));

      while Cur /= No_Element loop
         Old_Name := Element(Cur);

         if Index(Old_Name, Bwaa.Util.U2s(Old_Pkg_Name), 1) = 1 and
            not (Index(Old_Name, Bwaa.Util.U2s(New_Pkg_Name), 1) = 1) and
            (Length(Old_Name) <= Length(Old_Pkg_Name) or else
            Element(Old_Name, Length(Old_Pkg_Name) + 1) = '.') then

            -- extract the variable name (beware of records!)
            declare
               use Bwaa.Util;
               Old_Partition: Unbounded_String_Array := Split(Old_Name);
               Last: Natural := Old_Partition'Last - 1;
            begin
               while not Contains(Pa, Merge(Old_Partition, Last)) loop
                  if Last < 1 then
                     raise Package_Not_Found with "Unable to determine package portion of variable name " & Bwaa.Util.U2s(Old_Name);
                  end if;
                  Last := Last - 1;
               end loop;
               Variable_Name := Merge(Old_Partition, Last + 1, Old_Partition'Last);
            end;

            New_Name := Replace_Slice(Old_Name, 1, Length(Old_Pkg_Name), Bwaa.Util.U2s(New_Pkg_Name));
            Old_Prefix := Delete(Old_Name, Length(Old_Name) - Length(Variable_Name), Length(Old_Name));
            New_Prefix := Delete(New_Name, Length(New_Name) - Length(Variable_Name), Length(New_Name));

            --  Put_Line(Max, "[Bwaa.Aliasing] Extracting variable name:" &
--                         " old name: " & Bwaa.Util.U2s(Old_Name) &
--                         " new name: " & Bwaa.Util.U2s(New_Name) &
--                         " variable: " & Bwaa.Util.U2s(Variable_Name) &
--                         " old package: " & Bwaa.Util.U2s(Old_Prefix) &
--                         " new package: " & Bwaa.Util.U2s(New_Prefix));

            Circular_Package_Names(Aliases, Old_Prefix, New_Prefix, Circular);
            if Circular then
               Put_Line(Max, "[Bwaa.Aliasing] Skipped circular variable aliases: " & Bwaa.Util.U2s(Old_Name) & " => " & Bwaa.Util.U2s(New_Name));
            else
               Put_Line(Max, "[Bwaa.Aliasing] Updating variable alias " & Bwaa.Util.U2s(Old_Name) & " with " & Bwaa.Util.U2s(Old_Pkg_Name) & " => " & Bwaa.Util.U2s(New_Pkg_Name));
               Add_Variable_Alias(Aliases, Old_Name, New_Name);
            end if;
         end if;
         Next(Cur);
      end loop;
   end Update_Variable_Aliases;

end Bwaa.Aliasing;
