-- $Id: bwaa-analysis.adb 186 2009-07-28 06:38:38Z georg $
---------------------------------------------------------------------------
--                                                                       --
--                   BWAA - Busy Wait Analyser for Ada                   --
--      BWAA is a program that finds busy waiting in Ada programs.       --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--                 Copyright (c) 2009 Georg Kienesberger                 --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of BWAA.                                            --
--                                                                       --
-- BWAA is free software; you can redistribute it and/or modify          --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- BWAA is distributed in the hope that it will be useful,               --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Containers;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Ada.Characters.Handling; -- TODO: check

with Asis.Elements; -- TODO: check

with Ast2Cfg.Cfgs;
with Ast2Cfg.Flow_Types; use Ast2Cfg.Flow_Types; -- TODO: check
with Ast2Cfg.Nodes; use Ast2Cfg.Nodes; -- TODO: check
with Ast2Cfg.Id; use Ast2Cfg.Id; -- TODO: check
with Ast2Cfg.Utils;
use type Ast2Cfg.Long_Long_Natural;

with Bwaa.Traversal;
with Bwaa.Util; use Bwaa.Util;
with Bwaa.Graph; use Bwaa.Graph;
use type Vertex_Set.Cursor;
with Bwaa.Analysis_Types; use Bwaa.Analysis_Types;
with Bwaa.Analysis_Types.Variable_Set; use Bwaa.Analysis_Types.Variable_Set;
use type Bwaa.Analysis_Types.Variable_Set.Cursor;
with Bwaa.Analysis_Types.Container_Arrays; use Bwaa.Analysis_Types.Container_Arrays;
with Bwaa.Output; use Bwaa.Output;
with Bwaa.Union_Find;
with Bwaa.Tarjan;
with Bwaa.Options;
with Bwaa.Queue;
with Bwaa.Report;
use Bwaa;

use type Vertex_Map.Cursor;

package body Bwaa.Analysis is

   procedure Analyse (Ast2Cfg_Cfg: Ast2Cfg.Cfgs.Cfg_Object;
                      Specs: in Specification_Map.Map;
                      Visi_Info: in Visibility_Info.Info) is

      Bwaa_Cfg: Bwaa.Graph.Cfg(Ast2Cfg.Cfgs.Get_Node_Sum(Ast2Cfg_Cfg));
      BE_Cur: Vertex_Map.Cursor;
      LB_Cur, TS_Cur: Vertex_Set.Cursor;
      Cand_Cur: Variable_Set.Cursor;
   begin

      -- clean up from previous calls:
      Bwaa.Aliasing.Clear(Aliases);

      Put_Line(Verbose, "[Bwaa.Dfs] Analysing " & Ast2Cfg.Cfgs.Get_Name(Ast2Cfg_Cfg));

      Init(Bwaa_Cfg, Ast2Cfg_Cfg);

      Put_Line(Max, "----------Node-Mapping---------");

      for I in Bwaa_Cfg.Dfs_Sorted'Range loop
         Put_Line(Max, To_String(Get_Id(Node_Object(Bwaa_Cfg.Dfs_Sorted(I).all))) & " --> " &
                    Ast2Cfg.Long_Long_Natural'Image(Id_To_Dfs_Num.Element(Bwaa_Cfg.Id_Dfs,
                                                                          Get_Id(Node_Object(Bwaa_Cfg.Dfs_Sorted(I).all)))));
      end loop;

      Put_Line(Max, "-------------Succs-------------");

      for I in Bwaa_Cfg.Succs'Range loop
         Put(Max, To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(I).all)) & ": ");
         if Bwaa_Cfg.Succs(I) /= null then
            for J in Bwaa_Cfg.Succs(I)'Range loop
               Put(Max, To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.Succs(I)(J)).all)) & ", ");
            end loop;
         end if;
         New_Line(Max);
      end loop;

      Put_Line(Max, "-------------Preds-------------");

      for I in Bwaa_Cfg.Preds'Range loop
         Put(Max, To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(I).all)) & ": ");
         if Bwaa_Cfg.Preds(I) /= null then
            for J in Bwaa_Cfg.Preds(I)'Range loop
               Put(Max, To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.Preds(I)(J)).all)) & ", ");
            end loop;
         end if;
         New_Line(Max);
      end loop;

      Put_Line(Debug, "-----------Backedges-----------");

      BE_Cur := Vertex_Map.First(Bwaa_Cfg.Backedges);
      while BE_Cur /= Vertex_Map.No_Element loop
         Put_Line(Debug, "BE: " & To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Vertex_Map.Key(BE_Cur)).all)) &
                    " --> " & To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Vertex_Map.Element(BE_Cur)).all)));
         Vertex_Map.Next(BE_Cur);
      end loop;

      declare
         package T is new Tarjan(Bwaa_Cfg.Node_Count);
      begin
         T.Build_Loop_Forest(Bwaa_Cfg);
      end;

      Put_Line(Debug, "----------Loop Forest----------");
      Put_Line(Debug, "Loops:");
      for I in Bwaa_Cfg.L_Forest.Loops'Range loop
         Put(Debug, To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.L_Forest.Loops(I).Header).all)));
         Put(Debug, ": ");
         LB_Cur := Vertex_Set.First(Bwaa_Cfg.L_Forest.Loops(I).B0dy);
         while LB_Cur /= Vertex_Set.No_Element loop
            Put(Debug, To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Vertex_Set.Element(LB_Cur)).all)) & ", ");
            Vertex_Set.Next(LB_Cur);
         end loop;
         New_Line(Debug);
      end loop;
      New_Line(Debug);
      Put_Line(Debug, "Loop Preds:");
      for I in Bwaa_Cfg.L_Forest.Preds'Range loop
         Put(Debug, To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.L_Forest.Loops(I).Header).all)));
         Put(Debug, ": ");
         if Bwaa_Cfg.L_Forest.Preds(I) /= 0 then
            Put(Debug, To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.L_Forest.Loops(Bwaa_Cfg.L_Forest.Preds(I)).Header).all)));
         end if;
         New_Line(Debug);
      end loop;

      Find_Termination_Statements(Bwaa_Cfg);

      Put_Line(Debug, "-----Termination Statements----");

      for I in Bwaa_Cfg.L_Forest.Loops'Range loop
         Put(Debug, To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.L_Forest.Loops(I).Header).all)));
         Put(Debug, ": ");

         if Bwaa_Cfg.L_Forest.Loops(I).Terms /= null then
            for J in Bwaa_Cfg.L_Forest.Loops(I).Terms'Range loop
               Put(Debug, To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.L_Forest.Loops(I).Terms(J)).all)) & ", ");
            end loop;
         end if;
         New_Line(Debug);
      end loop;

      Aliasing.Collect(Aliases, Ast2Cfg_Cfg, Specs, Visi_Info);

      Put_Line(Debug, "------Collected Aliases-----");
      Put(Debug, Aliasing.Debug_Image(Aliases'Access));

      Find_Busy_Wait_Candidates(Bwaa_Cfg);
      if Options.Indirect_Candidates then
         Find_Indirect_Candidates(Bwaa_Cfg);
      end if;

      Put_Line(Debug, "------Busy Wait Candidates-----");

      for I in Bwaa_Cfg.L_Forest.Loops'Range loop
         if Bwaa_Cfg.L_Forest.Loops(I).Terms /= null then
            for J in Bwaa_Cfg.L_Forest.Loops(I).Terms'Range loop
               Put(Debug, To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.L_Forest.Loops(I).Terms(J)).all)));
               Put(Debug, ": ");

               Cand_Cur := Variable_Set.First(Bwaa_Cfg.L_Forest.Loops(I).Candidates(J));
               while Cand_Cur /= Variable_Set.No_Element loop
                  Put(Debug, Get_Name(Variable_Set.Element(Cand_Cur)) & ", ");
                  Variable_Set.Next(Cand_Cur);
               end loop;
               New_Line(Debug);
            end loop;
         end if;
      end loop;

      Find_Busy_Wait_Variables(Bwaa_Cfg);

      Put_Line(Debug, "------Busy Wait Variables-----");

      for I in Bwaa_Cfg.L_Forest.Loops'Range loop
         if Bwaa_Cfg.L_Forest.Loops(I).Terms /= null then
            for J in Bwaa_Cfg.L_Forest.Loops(I).Terms'Range loop
               Put(Debug, To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.L_Forest.Loops(I).Terms(J)).all)));
               Put(Debug, ": ");

               Cand_Cur := Variable_Set.First(Bwaa_Cfg.L_Forest.Loops(I).Candidates(J));
               while Cand_Cur /= Variable_Set.No_Element loop
                  if Is_Busy_Var(Variable_Set.Element(Cand_Cur)) then
                     Put(Debug, Get_Name(Variable_Set.Element(Cand_Cur)) & ", ");
                  end if;
                  Variable_Set.Next(Cand_Cur);
               end loop;
               New_Line(Debug);
            end loop;
         end if;
      end loop;

      New_Line(Debug);

      Report.Output(Bwaa_Cfg);

      Clear(Bwaa_Cfg);
   end Analyse;

   procedure Find_Termination_Statements (Bwaa_Cfg: in out Bwaa.Graph.Cfg) is
      Terms: Vertex_Set.Set;
      Cur: Vertex_Set.Cursor;
      Elem: Ast2Cfg.Long_Long_Natural;
      Header_Node: Ast2Cfg.Nodes.Node_Class_Ptr := null;
      use type Asis.Statement_Kinds;
   begin

      for I in Bwaa_Cfg.L_Forest.Loops'Range loop

         -- clear previously found termination statements
         Vertex_Set.Clear(Terms);

         -- ignore for loops for now, since they cannot be busy wait loops
         Header_Node := Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.L_Forest.Loops(I).Header);
         if (not (Header_Node.all in Finite_Loop_Node_Object'Class)) or else -- an ASIS query is more effort
            (not (Asis.Elements.Statement_Kind(Get_Element(Header_Node.all)) = Asis.A_For_Loop_Statement)) then

            -- check the header
            for J in Bwaa_Cfg.Succs(Bwaa_Cfg.L_Forest.Loops(I).Header)'Range loop
               if not Vertex_Set.Contains(Bwaa_Cfg.L_Forest.Loops(I).B0dy,
                                          Bwaa_Cfg.Succs(Bwaa_Cfg.L_Forest.Loops(I).Header)(J)) then
                  Vertex_Set.Insert(Terms, Bwaa_Cfg.L_Forest.Loops(I).Header);
                  exit;
               end if;
            end loop;

            -- check every node from the body
            Cur := Vertex_Set.First(Bwaa_Cfg.L_Forest.Loops(I).B0dy);
            while Cur /= Vertex_Set.No_Element loop
               Elem := Vertex_Set.Element(Cur);
               for J in Bwaa_Cfg.Succs(Elem)'Range loop
                  if not Vertex_Set.Contains(Bwaa_Cfg.L_Forest.Loops(I).B0dy, Bwaa_Cfg.Succs(Elem)(J)) and
                    Bwaa_Cfg.Succs(Elem)(J) /= Bwaa_Cfg.L_Forest.Loops(I).Header then
                     Vertex_Set.Insert(Terms, Elem);
                     exit;
                  end if;
               end loop;
               Vertex_Set.Next(Cur);
            end loop;

            -- save the found termination statements in the loop
            if not Vertex_Set.Is_Empty(Terms) then
               declare
                  J: Ast2Cfg.Long_Long_Natural := 1;
               begin
                  -- reserve space for the found termination statements
                  Bwaa_Cfg.L_Forest.Loops(I).Terms := new Vertex_Array(1 .. Ast2Cfg.Long_Long_Natural(Vertex_Set.Length(Terms)));

                  -- copy the found termination statements
                  Cur := Vertex_Set.First(Terms);
                  while Cur /= Vertex_Set.No_Element loop
                     Bwaa_Cfg.L_Forest.Loops(I).Terms(J) := Vertex_Set.Element(Cur);
                     J := J + 1;
                     Vertex_Set.Next(Cur);
                  end loop;
               end;
            end if;
         end if;
      end loop;
   end Find_Termination_Statements;

   -- check whether the variable or an equivalent name is already in the set
   function Already_In_Variable_Set (S: in Variable_Set.Set; Name: in Unbounded_String) return Boolean is
      Cur_Var: Cursor := No_Element;
   begin

      Cur_Var := First(S);
      while Cur_Var /= No_Element loop
         if Bwaa.Aliasing.Equal_Variable_Names(Aliases'Access, Get_Name(Element(Cur_Var)), Name) then
            return True;
         end if;
         Next(Cur_Var);
      end loop;

      return False;
   end Already_In_Variable_Set;

   procedure Harvest_Param_Tree (Cur: in Node_Class_Ptr; Variables: in out Variable_Set.Set) is
      Succs: Flow_Class_Ptr_List.Object;
      Succ: Node_Class_Ptr := null;
      Var_Obj: Variable;
      Var_Name: Unbounded_String := Null_Unbounded_String;
   begin

      if Cur = null then
         return;
      end if;

      Succs := Get_Succs(Cur.all);
      if Flow_Class_Ptr_List.Count(Succs) = 0 then
         if Cur.all in Param_Node_Object'Class then
            Var_Name := S2u(Get_Var_Name(Param_Node_Object(Cur.all)));
            if Var_Name /= "" then

               if Already_In_Variable_Set(Variables, Var_Name) then
                  return;
               end if;

               -- found a new variable, insert it
               Init(Var_Obj, Var_Name);
               Variable_Set.Include(Variables, Var_Obj);
            end if;
         else
            Put_Line(Debug, "[Bwaa.Analysis] Leaf of parameter tree is not in Param_Node_Object'Class: " & To_String(Get_Id(Cur.all)));
         end if;
      else
         Flow_Class_Ptr_List.Reset(Succs);
         while Flow_Class_Ptr_List.Has_Next(Succs) loop
            Flow_Class_Ptr_List.Get_Next(Succs, Flow_Class_Ptr(Succ));
            Harvest_Param_Tree(Succ, Variables);
         end loop;
      end if;

   end Harvest_Param_Tree;

   procedure Find_Busy_Wait_Candidates (Bwaa_Cfg: in out Bwaa.Graph.Cfg) is
      Node: Node_Class_Ptr := null;
      Idents: Ast2Cfg.Utils.UStr_List_List.Object;
      Ident: Ast2Cfg.Utils.UStr_List.Object;
      Succ: Ast2Cfg.Long_Long_Natural;
   begin

      for I in Bwaa_Cfg.L_Forest.Loops'Range loop

         -- if there are termination statements
         if Bwaa_Cfg.L_Forest.Loops(I).Terms /= null then

            -- we will save a set of variables for each termination statement
            Bwaa_Cfg.L_Forest.Loops(I).Candidates := new Variable_Set_Array(Bwaa_Cfg.L_Forest.Loops(I).Terms'Range);

            -- for each termination statement
            for J in Bwaa_Cfg.L_Forest.Loops(I).Terms'Range loop

               Node := Bwaa_Cfg.Dfs_Sorted((Bwaa_Cfg.L_Forest.Loops(I).Terms(J)));

               -- branch nodes need special treatment
               if Node.all in Branch_Node_Object'Class then
                  Idents := Get_Idents(Node.all);

                  if Ast2Cfg.Utils.UStr_List_List.Count(Idents) > 0 then
                     -- we encountered a case statement

                     if Get_RHS_Tree(Node.all) = null then
                        -- no param tree, so add the idents to the candidates
                        Ast2Cfg.Utils.UStr_List_List.Reset(Idents);
                        while Ast2Cfg.Utils.UStr_List_List.Has_Next(Idents) loop
                           Ast2Cfg.Utils.UStr_List_List.Get_Next(Idents, Ident);
                           declare
                              Var_Obj: Variable;
                           begin
                              Init(Var_Obj, Unbounded_String'(Ast2Cfg.Utils.Merge_UStr_List(Ident, ".")));
                              Variable_Set.Include(Bwaa_Cfg.L_Forest.Loops(I).Candidates(J), Var_Obj);
                           end;
                        end loop;
                     else
                        -- param tree, collect variables there
                        Harvest_Param_Tree(Get_RHS_Tree(Node.all),
                                           Bwaa_Cfg.L_Forest.Loops(I).Candidates(J));
                     end if;

                  else
                     -- if statement or case
                     -- we have to check every successor outside the loop

                     for K in Bwaa_Cfg.Succs(Bwaa_Cfg.L_Forest.Loops(I).Terms(J))'Range loop
                        Succ := Bwaa_Cfg.Succs(Bwaa_Cfg.L_Forest.Loops(I).Terms(J))(K);

                        if (not Vertex_Set.Contains(Bwaa_Cfg.L_Forest.Loops(I).B0dy, Succ)) and
                          (Succ /= Bwaa_Cfg.L_Forest.Loops(I).Header) then

                           -- collect the variables
                           Harvest_Param_Tree(Get_RHS_Tree(Bwaa_Cfg.Dfs_Sorted(Succ).all),
                                              Bwaa_Cfg.L_Forest.Loops(I).Candidates(J));
                        end if;
                     end loop;

                  end if;


               else

                  if Get_RHS_Tree(Node.all) = null then
                     -- no param tree, so add the idents to the candidates
                     Idents := Get_Idents(Node.all);
                     Ast2Cfg.Utils.UStr_List_List.Reset(Idents);
                     while Ast2Cfg.Utils.UStr_List_List.Has_Next(Idents) loop
                        Ast2Cfg.Utils.UStr_List_List.Get_Next(Idents, Ident);
                        declare
                           Var_Obj: Variable;
                        begin
                           Init(Var_Obj, Unbounded_String'(Ast2Cfg.Utils.Merge_UStr_List(Ident, ".")));
                           Variable_Set.Include(Bwaa_Cfg.L_Forest.Loops(I).Candidates(J), Var_Obj);
                        end;
                     end loop;
                  else
                     -- collect the variables
                     Harvest_Param_Tree(Get_RHS_Tree(Node.all),
                                        Bwaa_Cfg.L_Forest.Loops(I).Candidates(J));
                  end if;
               end if;
            end loop;
         end if;
      end loop;
   end Find_Busy_Wait_Candidates;

   procedure Find_Indirect_Candidates (Bwaa_Cfg: in out Bwaa.Graph.Cfg) is

      procedure Collect_Indirects (Node: in Node_Class_Ptr;
                                   Var: in Variable;
                                   Worklist: in out Variable_Set.Set;
                                   Candidates: in out Variable_Set.Set) is
         use type Ada.Containers.Count_Type;
         Left_Vars, Right_Vars, New_Candidates: Variable_Set.Set;
         Cur: Variable_Set.Cursor;
         Direct_Candidate_Name: Unbounded_String := Null_Unbounded_String;
         Cur_Var: Variable;
      begin
         if Node.all not in Assign_Node_Object'Class then
            return;
         end if;

         -- get the variable on the left hand side
         Harvest_Param_Tree(Get_LHS_Tree(Assign_Node_Object(Node.all)), Left_Vars);

         -- did we get more than one variable on the left hand side?
         if Variable_Set.Length(Left_Vars) /= 1 then
            raise Bwaa_Exception with "Analysis: More than one variable on the left hand side!";
         end if;

         -- is this an assignment to the variable in question?
         Cur := Variable_Set.First(Left_Vars);
         if not Bwaa.Aliasing.Equal_Variable_Names(Aliases'Access, Get_Name(Var), Get_Name(Variable_Set.Element(Cur))) then
            return;
         end if;

         -- find all variables on the right hand side
         Harvest_Param_Tree(Get_RHS_Tree(Node.all), Right_Vars);

         -- find the candidates that are new
         Cur := Variable_Set.First(Right_Vars);
         while Cur /= Variable_Set.No_Element loop
            Cur_Var := Variable_Set.Element(Cur);
            if not Already_In_Variable_Set(Candidates, Get_Name(Cur_Var)) then
               Variable_Set.Insert(New_Candidates, Cur_Var);
            end if;

            Variable_Set.Next(Cur);
         end loop;

         -- set the name of the direct candidate of each indirect one
         if Unbounded_String'(Get_Direct_Candidate_Name(Var)) = Null_Unbounded_String then
            Direct_Candidate_Name := Get_Name(Var);
         else
            Direct_Candidate_Name := Get_Direct_Candidate_Name(Var);
         end if;
         Cur := Variable_Set.First(New_Candidates);
         while Cur /= Variable_Set.No_Element loop
            Cur_Var := Variable_Set.Element(Cur);
            Set_Direct_Candidate_Name(Cur_Var, Direct_Candidate_Name);
            Variable_Set.Replace_Element(New_Candidates, Cur, Cur_Var);
            Variable_Set.Next(Cur);
         end loop;

         -- add the newly found candidates to the worklist
         Worklist := Variable_Set.Union(Worklist, New_Candidates);

         -- add the found indirect candidates to the candidates
         Candidates := Variable_Set.Union(Candidates, New_Candidates); -- TODO: check: -- Right_Vars);

      end Collect_Indirects;

      LB_Cur: Vertex_Set.Cursor;
      Var_Cur: Variable_Set.Cursor;
      Var: Variable;
      Worklist: Variable_Set.Set;
   begin

      for I in Bwaa_Cfg.L_Forest.Loops'Range loop

         -- if there are termination statements
         if Bwaa_Cfg.L_Forest.Loops(I).Terms /= null then

            -- for each termination statement
            for J in Bwaa_Cfg.L_Forest.Loops(I).Terms'Range loop

               -- initialise worklist with the direct candidates
               Worklist := Bwaa_Cfg.L_Forest.Loops(I).Candidates(J);

               Var_Cur := Variable_Set.First(Worklist);
               while Var_Cur /= Variable_Set.No_Element loop

                  -- pick an element from the worklist
                  Var := Variable_Set.Element(Var_Cur);
                  Variable_Set.Delete(Worklist, Var_Cur);

                  -- search the header for indirect candidates,
                  -- add them to the candidates and the worklist
                  Collect_Indirects(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.L_Forest.Loops(I).Header),
                                    Var,
                                    Worklist,
                                    Bwaa_Cfg.L_Forest.Loops(I).Candidates(J));

                  -- search the body for indirect candidates,
                  -- add them to the candidates and the worklist
                  LB_Cur := Vertex_Set.First(Bwaa_Cfg.L_Forest.Loops(I).B0dy);
                  while LB_Cur /= Vertex_Set.No_Element loop
                     Collect_Indirects(Bwaa_Cfg.Dfs_Sorted(Vertex_Set.Element(LB_Cur)),
                                       Var,
                                       Worklist,
                                       Bwaa_Cfg.L_Forest.Loops(I).Candidates(J));

                     Vertex_Set.Next(LB_Cur);
                  end loop;

                  Var_Cur := Variable_Set.First(Worklist);
               end loop;

            end loop;
         end if;
      end loop;
   end Find_Indirect_Candidates;

   function Defines (Node: in Ast2Cfg.Nodes.Node_Class_Ptr; Var: in Variable) return Boolean is
      Ptree_Root : Ast2Cfg.Nodes.Node_Class_Ptr := null;
   begin
      if not (Node.all in Ast2Cfg.Nodes.Assign_Node_Object) then
         return False;
      end if;

      Ptree_Root := Ast2Cfg.Nodes.Get_LHS_Tree(Assign_Node_Object(Node.all));

      if Ptree_Root.all in Param_Node_Object then
         if Bwaa.Aliasing.Equal_Variable_Names(Aliases'Access,
                                               Get_Name(Var),
                                               S2u(Ast2Cfg.Nodes.Get_Var_Name(Param_Node_Object(Ptree_Root.all)))) then
            return True;
         end if;
      end if;

      return False;
   end Defines;

   procedure Busy_Var_Check (Bwaa_Cfg: in Bwaa.Graph.Cfg;
                             L00p: in Bwaa.Graph.L00p;
                             Termination_Statement: in Ast2Cfg.Long_Long_Natural;
                             BW_Candidate: in out Variable) is

      package Queue is new Bwaa.Queue(Ast2Cfg.Long_Long_Natural);
      use Queue;

      Q: Queue.Object;
      Visited: Vertex_Set.Set;
      use Vertex_Set;
      Cur: Ast2Cfg.Long_Long_Natural;
   begin

      Put(Q, Termination_Statement);
      Put_Line(Debug, "TS: " & To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Termination_Statement).all)));
      while not Is_Empty(Q) loop
         Get(Q, Cur);

         for J in Bwaa_Cfg.Preds(Cur)'Range loop

            if Bwaa_Cfg.Preds(Cur)(J) = Termination_Statement then
               Put_Line(Debug, "Stop");
               Set_Busy_Var(BW_Candidate);
               return;
            end if;

            if Contains(L00p.B0dy, Bwaa_Cfg.Preds(Cur)(J)) or else Bwaa_Cfg.Preds(Cur)(J) = L00p.Header then
               if not Contains(Visited, Bwaa_Cfg.Preds(Cur)(J)) and then
                  not Defines(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.Preds(Cur)(J)), BW_Candidate) then
                  Put(Q, Bwaa_Cfg.Preds(Cur)(J));
                  Insert(Visited, Bwaa_Cfg.Preds(Cur)(J));
                  Put_Line(Debug, To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.Preds(Cur)(J)).all)));
               else
                  Put_Line(Debug, "Skipped: " & To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.Preds(Cur)(J)).all)));
               end if;
            elsif Cur /= L00p.Header then
               raise Analysis_Exception with "[Is_Busy_Var] Node in loop body with predecessor outside of loop encountered.";
            else
               Put_Line(Debug, "Skipped(H): " & To_String(Get_Id(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.Preds(Cur)(J)).all)));
            end if;
         end loop;
      end loop;
   end Busy_Var_Check;

   procedure Find_Busy_Wait_Variables (Bwaa_Cfg: in out Bwaa.Graph.Cfg) is
      Cand_Cur: Variable_Set.Cursor;
      Cand: Variable;
   begin

      for I in Bwaa_Cfg.L_Forest.Loops'Range loop
         if Bwaa_Cfg.L_Forest.Loops(I).Terms /= null then
            for J in Bwaa_Cfg.L_Forest.Loops(I).Terms'Range loop

               Cand_Cur := Variable_Set.First(Bwaa_Cfg.L_Forest.Loops(I).Candidates(J));
               while Cand_Cur /= Variable_Set.No_Element loop
                  Cand := Variable_Set.Element(Cand_Cur);

                  Busy_Var_Check(Bwaa_Cfg,
                                 Bwaa_Cfg.L_Forest.Loops(I),
                                 Bwaa_Cfg.L_Forest.Loops(I).Terms(J),
                                 Cand);

                  Variable_Set.Replace(Bwaa_Cfg.L_Forest.Loops(I).Candidates(J), Cand);

                  Variable_Set.Next(Cand_Cur);
               end loop;
            end loop;
         end if;
      end loop;

   end Find_Busy_Wait_Variables;

end Bwaa.Analysis;
