-- $Id: bwaa-graph.adb 64 2009-04-14 17:02:27Z georg $
---------------------------------------------------------------------------
--                                                                       --
--                   BWAA - Busy Wait Analyser for Ada                   --
--      BWAA is a program that finds busy waiting in Ada programs.       --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--                 Copyright (c) 2009 Georg Kienesberger                 --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of BWAA.                                            --
--                                                                       --
-- BWAA is free software; you can redistribute it and/or modify          --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- BWAA is distributed in the hope that it will be useful,               --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Containers;

with Ast2Cfg.Flow_Types; use Ast2Cfg.Flow_Types;
with Ast2Cfg.Nodes; use Ast2Cfg.Nodes;

with Bwaa.Traversal;
use Bwaa;

package body Bwaa.Graph is

   procedure Init (Bwaa_Cfg: out Cfg; Ast2Cfg_Cfg: in Ast2Cfg.Cfgs.Cfg_Object) is
      Succs, Preds            : Flow_Class_Ptr_List.Object;
      Succ, Pred              : Node_Class_Ptr := null;
      Succs_Array, Preds_Array: Vertex_Array_Ptr := null;
      J: Ast2Cfg.Long_Long_Natural := 1;
   begin

      Id_To_Dfs_Num.Reserve_Capacity(Bwaa_Cfg.Id_Dfs, Ada.Containers.Count_Type(Ast2Cfg.Cfgs.Get_Node_Sum(Ast2Cfg_Cfg)));

      -- Fill Bwaa_Cfg.Dfs_Sorted and Bwaa_Cfg.Id_Dfs, find the backedges
      Traversal.Dfs(Ast2Cfg_Cfg, Bwaa_Cfg);

      -- build the adjacency lists (succs, preds)
      for I in Bwaa_Cfg.Dfs_Sorted'Range loop
         -- add the successors of the current node
         Succs := Get_Succs(Bwaa_Cfg.Dfs_Sorted(I).all);
         Flow_Class_Ptr_List.Reset(Succs);
         if Flow_Class_Ptr_List.Count(Succs) > 0 then
            Succs_Array := new Vertex_Array(1 .. Flow_Class_Ptr_List.Count(Succs));
            J := 1;
            while Flow_Class_Ptr_List.Has_Next(Succs) loop
               Flow_Class_Ptr_List.Get_Next(Succs, Flow_Class_Ptr(Succ));
               Succs_Array(J) := Id_To_Dfs_Num.Element(Bwaa_Cfg.Id_Dfs, Get_Id(Node_Object(Succ.all)));
               J := J + 1;
            end loop;
         else
            Succs_Array := null;
         end if;
         Bwaa_Cfg.Succs(I) := Succs_Array;

         -- add the predecessors of the current node
         Preds := Get_Preds(Bwaa_Cfg.Dfs_Sorted(I).all);
         Flow_Class_Ptr_List.Reset(Preds);
         if Flow_Class_Ptr_List.Count(Preds) > 0 then
            Preds_Array := new Vertex_Array(1 .. Flow_Class_Ptr_List.Count(Preds));
            J := 1;
            while Flow_Class_Ptr_List.Has_Next(Preds) loop
               Flow_Class_Ptr_List.Get_Next(Preds, Flow_Class_Ptr(Pred));
               Preds_Array(J) := Id_To_Dfs_Num.Element(Bwaa_Cfg.Id_Dfs, Get_Id(Node_Object(Pred.all)));
               J := J + 1;
            end loop;
         else
            Preds_Array := null;
         end if;
         Bwaa_Cfg.Preds(I) := Preds_Array;
      end loop;

   end Init;

   procedure Clear (L: in out L00p) is
   begin
      Free(L.Terms);
   end Clear;

   procedure Init (Forest: in out Loop_Forest) is
   begin
      Forest.Preds := (others => 0);
   end Init;

   procedure Clear (A: in out Adjacency_Array) is
   begin
      for I in A'Range loop
         Free(A(I));
      end loop;
   end Clear;

   procedure Clear (Forest: in out Loop_Forest) is
   begin
      for I in Forest.Loops'Range loop
         Vertex_Set.Clear(Forest.Loops(I).B0dy);
         Free(Forest.Loops(I).Terms);
      end loop;
   end Clear;

   procedure Free (Forest_Ptr: in out Loop_Forest_Ptr) is
      procedure Destroy is new Ada.Unchecked_Deallocation(Loop_Forest, Loop_Forest_Ptr);
   begin
      Clear(Forest_Ptr.all);
      Destroy(Forest_Ptr);
   end Free;

   procedure Clear (Bwaa_Cfg: in out Cfg) is
   begin
      Clear(Bwaa_Cfg.Succs);
      Clear(Bwaa_Cfg.Preds);
      Id_To_Dfs_Num.Clear(Bwaa_Cfg.Id_Dfs);
      Vertex_Map.Clear(Bwaa_Cfg.Backedges);
      Vertex_Set.Clear(Bwaa_Cfg.Headers);
      Free(Bwaa_Cfg.L_Forest);
   end Clear;

end Bwaa.Graph;
