-- $Id: bwaa-graph.ads 64 2009-04-14 17:02:27Z georg $
---------------------------------------------------------------------------
--                                                                       --
--                   BWAA - Busy Wait Analyser for Ada                   --
--      BWAA is a program that finds busy waiting in Ada programs.       --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--                 Copyright (c) 2009 Georg Kienesberger                 --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of BWAA.                                            --
--                                                                       --
-- BWAA is free software; you can redistribute it and/or modify          --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- BWAA is distributed in the hope that it will be useful,               --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;
with Ada.Containers.Hashed_Maps;
with Ada.Containers.Hashed_Sets;

with Ast2Cfg;
with Ast2Cfg.Nodes;
with Ast2Cfg.Cfgs;

with Bwaa.Util; use Bwaa.Util;
with Bwaa.Analysis_Types.Container_Arrays; use Bwaa.Analysis_Types.Container_Arrays;

use type Ast2Cfg.Long_Long_Natural;

package Bwaa.Graph is

   type Node_Array is array (Ast2Cfg.Long_Long_Natural range <>) of Ast2Cfg.Nodes.Node_Class_Ptr;

   type Vertex_Array is array (Ast2Cfg.Long_Long_Natural range <>) of Ast2Cfg.Long_Long_Natural;
   type Vertex_Array_Ptr is access Vertex_Array;
   procedure Free is new Ada.Unchecked_Deallocation(Vertex_Array, Vertex_Array_Ptr);

   package Vertex_Map is new Ada.Containers.Hashed_Maps (Key_Type => Ast2Cfg.Long_Long_Natural,
                                                         Element_Type => Ast2Cfg.Long_Long_Natural,
                                                         Hash => Hash,
                                                         Equivalent_Keys => "=");

   package Vertex_Set is new Ada.Containers.Hashed_Sets (Element_Type => Ast2Cfg.Long_Long_Natural,
                                                         Hash => Hash,
                                                         Equivalent_Elements => "=");

   type Adjacency_Array is array (Ast2Cfg.Long_Long_Natural range <>) of Vertex_Array_Ptr;
   procedure Clear (A: in out Adjacency_Array);

   type L00p is
      record
         Header: Ast2Cfg.Long_Long_Natural; -- the dfs number of the loop header
         B0dy: Vertex_Set.Set; -- the dfs numbers of the nodes in the loop body
         Terms: Vertex_Array_Ptr := null; -- the termination statements
         Candidates: Variable_Set_Array_Ptr := null; -- the busy wait candidates
      end record;

   procedure Clear (L: in out L00p);

   type L00p_Array is array (Ast2Cfg.Long_Long_Natural range <>) of L00p;

   type Loop_Forest (Loop_Count: Ast2Cfg.Long_Long_Natural) is
      record
         Preds: Vertex_Array(1 .. Loop_Count); -- the predecessor of a loop, 0 if none
         Loops: L00p_Array(1 .. Loop_Count); -- the loops in reverse topological order
      end record;

   type Loop_Forest_Ptr is access Loop_Forest;

   procedure Init (Forest: in out Loop_Forest);
   procedure Clear (Forest: in out Loop_Forest);
   procedure Free (Forest_Ptr: in out Loop_Forest_Ptr);

   type Cfg (Node_Count: Ast2Cfg.Long_Long_Natural) is limited
      record
         Dfs_Sorted: Node_Array(1 .. Node_Count) := (others => null); -- maps dfs numbers to the node objects of Ast2Cfg
         Succs     : Adjacency_Array(1 .. Node_Count) := (others => null); -- the successors for each dfs number
         Preds     : Adjacency_Array(1 .. Node_Count) := (others => null); -- the predecessors for each dfs number
         Id_Dfs    : Id_To_Dfs_Num.Map; -- maps the Id of an Ast2Cfg node back to the dfs number
         Backedges : Vertex_Map.Map; -- the backedges (m,n) where m is the key of the map
         Headers   : Vertex_Set.Set; -- the set of loop headers (targets of backedges)
         L_Forest  : Loop_Forest_Ptr; -- the loop forest
      end record;

   procedure Init (Bwaa_Cfg: out Cfg; Ast2Cfg_Cfg: in Ast2Cfg.Cfgs.Cfg_Object);
   procedure Clear (Bwaa_Cfg: in out Cfg);

end Bwaa.Graph;
