-- $Id: bwaa-hashed_union_find.adb 101 2009-06-07 16:36:46Z georg $
---------------------------------------------------------------------------
--                                                                       --
--                   BWAA - Busy Wait Analyser for Ada                   --
--      BWAA is a program that finds busy waiting in Ada programs.       --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--                 Copyright (c) 2009 Georg Kienesberger                 --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of BWAA.                                            --
--                                                                       --
-- BWAA is free software; you can redistribute it and/or modify          --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- BWAA is distributed in the hope that it will be useful,               --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

package body Bwaa.Hashed_Union_Find is

   -- create a set containing only X
   procedure Makeset (S: in out Sets; X: in Element_Type) is
   begin

      Element_Map.Insert(S.Elements, X, X);

      declare
      begin
         Rank_Map.Insert(S.Ranks, X, 0);
      exception
         when Constraint_Error =>
            Element_Map.Delete(S.Elements, X);
            raise;
      end;
   end Makeset;

   -- finds the canonical element for the set containing X; uses path compression
   procedure Find (S: in out Sets; X: in Element_Type; Canonical: out Element_Type) is
      Root, Cur, Next: Element_Type;
   begin

      if not Contains(S, X) then
         raise Not_In_Any_Set_Exception;
      end if;

      Root := Element_Map.Element(S.Elements, X);
      while Root /= Element_Map.Element(S.Elements, Root) loop
         Root := Element_Map.Element(S.Elements, Root);
      end loop;

      -- path compression
      Next := X;
      while Next /= Root loop
         Cur := Next;
         Next := Element_Map.Element(S.Elements, Cur);
         Element_Map.Replace(S.Elements, Cur, Root);
      end loop;

      Canonical := Root;
   end Find;

   -- links X to Y, so that find on X and all its descendants returns
   -- the canonical element of the set containing Y
   procedure Link (S: in out Sets; X: in Element_Type; Y: in Element_Type) is
      Rank_X: Rank;
      Rank_Y: Rank;
   begin

      if X = Y then
         return;
      end if;

      if not Contains(S, X) or else
         not Contains(S, Y) then
         raise Not_In_Any_Set_Exception;
      end if;

      Rank_X := Rank_Map.Element(S.Ranks, X);
      Rank_Y := Rank_Map.Element(S.Ranks, Y);

      if Rank_X > Rank_Y then
         Element_Map.Replace(S.Elements, Y, X);
      elsif Rank_X < Rank_Y then
         Element_Map.Replace(S.Elements, X, Y);
      else
         Element_Map.Replace(S.Elements, X, Y);
         Rank_Map.Replace(S.Ranks, Y, Rank_Y + 1);
      end if;
   end Link;

   -- links the canonical elements of the sets containing X or Y, respectively
   procedure Union (S: in out Sets; X: in Element_Type; Y: in Element_Type) is
      Fx, Fy: Element_Type;
   begin

      if X = Y then
         return;
      end if;

      Find(S, X, Fx);
      Find(S, Y, Fy);
      Link(S, Fx, Fy);
   end Union;

   --  -- Result is true if both elements are in the same set -- TODO: remove
--     procedure In_Same_Set (S: in out Sets; X: in Element_Type; Y: in Element_Type; Result: out Boolean) is
--        Fx, Fy: Element_Type;
--     begin
--        Find(S, X, Fx);
--        Find(S, Y, Fy);
--        Result := Fx = Fy;
--     end In_Same_Set;

   -- return the number of all elements in all sets
   function Length (S: in Sets) return Ada.Containers.Count_Type is
   begin
      return Element_Map.Length(S.Elements);
   end Length;

   -- check whether the element is in one of the sets
   function Contains (S: in Sets; X: in Element_Type) return Boolean is
   begin
      return Element_Map.Contains(S.Elements, X);
   end Contains;

   -- get a cursor on the first element in the set containing the given element
   procedure First (S: in out Sets; X: in Element_Type; Cur: out Cursor) is
      use type Element_Map.Cursor;
      Canonical: Element_Type := Null_Element;
   begin
      Cur.Cur := Element_Map.First(S.Elements);
      Find(S, X, Cur.Canonical);

      while Cur.Cur /= Element_Map.No_Element loop
         Find(S, Element_Map.Element(Cur.Cur), Canonical);
         if Canonical = Cur.Canonical then
            return;
         end if;
         Element_Map.Next(Cur.Cur);
      end loop;

      Cur := No_Element;

   exception
      when Not_In_Any_Set_Exception =>
         Cur := No_Element;
   end First;

   -- get a cursor on the first of all elements
   function First (S: in Sets) return Cursor is
      Cur: Cursor;
   begin
      Cur.Cur := Element_Map.First(S.Elements);
      return Cur;
   end First;

   -- get a cursor on the next element, regardless of the set it is in
   function Next (Position: in Cursor) return Cursor is
      Cur: Cursor;
   begin
      if Position.Canonical /= Null_Element then
         raise Hashed_Union_Find_Exception with "Unexpected switch between cursor modes.";
      else
         Cur.Cur := Element_Map.Next(Position.Cur);
         return Cur;
      end if;
   end Next;

   -- change the cursor to point on the next element, regardless of the set it is in
   procedure Next (Position: in out Cursor) is
   begin
      if Position.Canonical /=  Null_Element then
         raise Hashed_Union_Find_Exception with "Unexpected switch between cursor modes.";
      else
         Element_Map.Next(Position.Cur);
      end if;

   end Next;

   -- change the cursor to point on the next element in the same set
   procedure Next (S: in out Sets; Position: in out Cursor) is
      use type Element_Map.Cursor;
      Cur: Cursor renames Position;
      Canonical: Element_Type := Null_Element;
   begin

      if Cur.Canonical /=  Null_Element then

         Element_Map.Next(Cur.Cur);
         while Cur.Cur /= Element_Map.No_Element loop
            Find(S, Element_Map.Element(Cur.Cur), Canonical);
            if Canonical = Cur.Canonical then
               return;
            end if;
            Element_Map.Next(Cur.Cur);
         end loop;

         Cur := No_Element;
      else
         raise Hashed_Union_Find_Exception with "Unexpected switch between cursor modes.";
      end if;

   end Next;

   -- check whether there is an element or No_Element at the cursor position
   function Has_Element (Position: Cursor) return Boolean is
   begin
      return Element_Map.Has_Element(Position.Cur);
   end Has_Element;

   -- return the element at the cursor position
   function Element (Position: in Cursor) return Element_Type is
   begin
      return Element_Map.Key(Position.Cur);
   end Element;

   -- return the parent of the element at the cursor position (for debugging purposes)
   function Parent (Position: in Cursor) return Element_Type is
   begin
      return Element_Map.Element(Position.Cur);
   end Parent;

   -- remove all elements
   procedure Clear (S: in out Sets) is
   begin
      Element_Map.Clear(S.Elements);
      Rank_Map.Clear(S.Ranks);
   end Clear;

end Bwaa.Hashed_Union_Find;
