-- $Id: bwaa-main.adb 186 2009-07-28 06:38:38Z georg $
---------------------------------------------------------------------------
--                                                                       --
--                   BWAA - Busy Wait Analyser for Ada                   --
--      BWAA is a program that finds busy waiting in Ada programs.       --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--                 Copyright (c) 2009 Georg Kienesberger                 --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of BWAA.                                            --
--                                                                       --
-- BWAA is free software; you can redistribute it and/or modify          --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- BWAA is distributed in the hope that it will be useful,               --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Exceptions;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Ada.Characters.Handling; use Ada.Characters.Handling;
with Ada.Containers; use type Ada.Containers.Count_Type; -- TODO: remove

with Ast2Cfg.Control;
with Ast2Cfg.Flow_World;
with Ast2Cfg.Output;
with Ast2Cfg.Pkgs; use Ast2Cfg.Pkgs;
with Ast2Cfg.Cfgs; use Ast2Cfg.Cfgs;
with Ast2Cfg.Flow_Types; use Ast2Cfg.Flow_Types;

with Bwaa.Options;
with Bwaa.Traversal;
with Bwaa.Analysis;
with Bwaa.Visibility; use Bwaa.Visibility;
with Bwaa.Visibility.Visibility_Info; use Bwaa.Visibility.Visibility_Info;
with Bwaa.Util;
with Bwaa.Output; use Bwaa.Output;

package body Bwaa.Main is
   procedure Run is
      World: Ast2Cfg.Flow_World.World_Object_Ptr;
      Specs: Bwaa.Util.Specification_Map.Map;
      Visi_Info: Visibility_Info.Info;

      Pkgs: Pkg_Class_Ptr_List.Object;
      Pkg:  Pkg_Class_Ptr := null;

      -- Find all Cfg_Objects in a given PKG/CFG tree and call Analysis.Analyse for each of them
      procedure Iterate_Cfgs (Flow: in Flow_Object'Class) is
         Flows: Flow_Class_Ptr_List.Object;
         Next_Flow: Flow_Class_Ptr := null;
      begin
         if (Flow in CFG_Object'Class) then
            Analysis.Analyse(CFG_Object(Flow), Specs, Visi_Info);
         elsif (Flow in Pkg_Object'Class) then
            null;
         else
            raise Bwaa_Exception with "Main: Wrong data structure encountered. Expected CFG_Object/Pkg_Object.";
         end if;

         Flows := Get_Succs(Flow);
         Flow_Class_Ptr_List.Reset(Flows);
         while Flow_Class_Ptr_List.Has_Next(Flows) loop
            Flow_Class_Ptr_List.Get_Next(Flows, Next_Flow);
            Iterate_Cfgs(Next_Flow.all);
         end loop;
      end Iterate_Cfgs;

   begin
      ---------------------
      -- Initializations --
      ---------------------
      case Bwaa.Options.Verbosity is
         when 0 =>
            Ast2Cfg.Output.Set_Level(Ast2Cfg.Output.None);
            Bwaa.Output.Set_Level(Bwaa.Output.Warning);
         when 1 =>
            Ast2Cfg.Output.Set_Level(Ast2Cfg.Output.None);
            Bwaa.Output.Set_Level(Bwaa.Output.PrettyMsg);
         when 2 =>
            Ast2Cfg.Output.Set_Level(Ast2Cfg.Output.Error);
            Bwaa.Output.Set_Level(Bwaa.Output.Verbose);
         when 3 =>
            Ast2Cfg.Output.Set_Level(Ast2Cfg.Output.Debug);
            Bwaa.Output.Set_Level(Bwaa.Output.Debug);
         when 4 =>
            Ast2Cfg.Output.Set_Level(Ast2Cfg.Output.Max);
            Bwaa.Output.Set_Level(Bwaa.Output.Max);
      end case;

      Ast2Cfg.Control.Init(To_Wide_String(To_String(Options.Context_Parameters)));

      -----------------------------------
      -- Fill the World with flow data --
      -----------------------------------
      World := Ast2Cfg.Control.Generate;

      ----------------------------
      -- Utilize the Flow World --
      ----------------------------
      Pkgs := Ast2Cfg.Flow_World.Get_Pkgs(World.all);

      -- collect all specifications
      Pkg_Class_Ptr_List.Reset(Pkgs);
      while Pkg_Class_Ptr_List.Has_Next(Pkgs) loop
         Pkg_Class_Ptr_List.Get_Next(Pkgs, Pkg);
         Bwaa.Traversal.Collect_Specs(Flow_Class_Ptr(Pkg), Specs, Null_Unbounded_String);
      end loop;

      -- build the visibility map
      Pkg_Class_Ptr_List.Reset(Pkgs);
      while Pkg_Class_Ptr_List.Has_Next(Pkgs) loop
         Pkg_Class_Ptr_List.Get_Next(Pkgs, Pkg);
         Bwaa.Traversal.Build_Visibility_Info(Flow_Class_Ptr(Pkg),
                                              Visi_Info,
                                              Visible_Set.Empty_Set,
                                              Visible_Set.Empty_Set,
                                              Specs);
      end loop;

      -- TODO: remove
      declare
         use type Visible_Set.Cursor;
         use type Visibility_Map.Cursor;
         Cur: Visibility_Map.Cursor := Visibility_Map.First(Visi_Info.Specs);
         Set: Visible_Set.Set;
         Cur_Set: Visible_Set.Cursor := Visible_Set.No_Element;
      begin
         Put_Line(Max, "Visibility_Map_Specs:");

         while Cur /= Visibility_Map.No_Element loop
            Set := Visibility_Map.Element(Cur);
            Cur_Set := Visible_Set.First(Set);
            Put(Max, Bwaa.Util.U2s(Visibility_Map.Key(Cur)) & ": ");

            while Cur_Set /= Visible_Set.No_Element loop
               Put(Max, Bwaa.Util.U2s(Bwaa.Visibility.Get_Full_Name(Visible_Set.Element(Cur_Set))) & " ");

               Visible_Set.Next(Cur_Set);
            end loop;

            New_Line(Max);

            Visibility_Map.Next(Cur);
         end loop;
      end;

      -- TODO: remove
      declare
         use type Visible_Set.Cursor;
         use type Visibility_Map.Cursor;
         Cur: Visibility_Map.Cursor := Visibility_Map.First(Visi_Info.Bodies);
         Set: Visible_Set.Set;
         Cur_Set: Visible_Set.Cursor := Visible_Set.No_Element;
      begin
         Put_Line(Max, "Visibility_Map_Bodies:");

         while Cur /= Visibility_Map.No_Element loop
            Set := Visibility_Map.Element(Cur);
            Cur_Set := Visible_Set.First(Set);
            Put(Max, Bwaa.Util.U2s(Visibility_Map.Key(Cur)) & ": ");

            while Cur_Set /= Visible_Set.No_Element loop
               Put(Max, Bwaa.Util.U2s(Bwaa.Visibility.Get_Full_Name(Visible_Set.Element(Cur_Set))) & " ");

               Visible_Set.Next(Cur_Set);
            end loop;

            New_Line(Max);

            Visibility_Map.Next(Cur);
         end loop;
      end;

      -- TODO: remove
      declare
         use type Visible_Set.Cursor;
         use type Visibility_Map.Cursor;
         Cur: Visibility_Map.Cursor := Visibility_Map.First(Visi_Info.Pkg_Specs);
         Set: Visible_Set.Set;
         Cur_Set: Visible_Set.Cursor := Visible_Set.No_Element;
      begin
         Put_Line(Max, "Pkg_Visibility_Map_Specs:");

         while Cur /= Visibility_Map.No_Element loop
            Set := Visibility_Map.Element(Cur);
            Cur_Set := Visible_Set.First(Set);
            Put(Max, Bwaa.Util.U2s(Visibility_Map.Key(Cur)) & ": ");

            while Cur_Set /= Visible_Set.No_Element loop
               Put(Max, Bwaa.Util.U2s(Bwaa.Visibility.Get_Full_Name(Visible_Set.Element(Cur_Set))) & " ");

               Visible_Set.Next(Cur_Set);
            end loop;

            New_Line(Max);

            Visibility_Map.Next(Cur);
         end loop;
      end;

      -- TODO: remove
      declare
         use type Visible_Set.Cursor;
         use type Visibility_Map.Cursor;
         Cur: Visibility_Map.Cursor := Visibility_Map.First(Visi_Info.Pkg_Bodies);
         Set: Visible_Set.Set;
         Cur_Set: Visible_Set.Cursor := Visible_Set.No_Element;
      begin
         Put_Line(Max, "Pkg_Visibility_Map_Bodies:");

         while Cur /= Visibility_Map.No_Element loop
            Set := Visibility_Map.Element(Cur);
            Cur_Set := Visible_Set.First(Set);
            Put(Max, Bwaa.Util.U2s(Visibility_Map.Key(Cur)) & ": ");

            while Cur_Set /= Visible_Set.No_Element loop
               Put(Max, Bwaa.Util.U2s(Bwaa.Visibility.Get_Full_Name(Visible_Set.Element(Cur_Set))) & " ");

               Visible_Set.Next(Cur_Set);
            end loop;

            New_Line(Max);

            Visibility_Map.Next(Cur);
         end loop;
      end;

      -- analyse each CFG in the Pkg/CFG tree
      Pkg_Class_Ptr_List.Reset(Pkgs);
      while Pkg_Class_Ptr_List.Has_Next(Pkgs) loop
         Pkg_Class_Ptr_List.Get_Next(Pkgs, Pkg);
         Put_Line(Debug, "[Bwaa.Main] At top-level package: " & Ast2Cfg.Flow_Types.Get_Name(Flow_Object(Pkg.all)));
         Iterate_Cfgs(Pkg.all);
      end loop;

      ------------------
      -- Finalization --
      ------------------
      Ast2Cfg.Control.Final;

   exception
      when Ex : others =>

         Put_Line(Error, "[Bwaa.Main] ERROR: " &
                  Ada.Exceptions.Exception_Name(Ex) &
                  ": " &
                  Ada.Exceptions.Exception_Information(Ex));

   end Run;
end Bwaa.Main;
