-- $Id: bwaa-report.adb 174 2009-07-25 13:58:44Z georg $
---------------------------------------------------------------------------
--                                                                       --
--                   BWAA - Busy Wait Analyser for Ada                   --
--      BWAA is a program that finds busy waiting in Ada programs.       --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--                 Copyright (c) 2009 Georg Kienesberger                 --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of BWAA.                                            --
--                                                                       --
-- BWAA is free software; you can redistribute it and/or modify          --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- BWAA is distributed in the hope that it will be useful,               --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Strings.Fixed; use Ada.Strings.Fixed;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;

with Asis.Elements;
with Asis.Compilation_Units;
with Asis.Text; use Asis.Text;

with Ast2Cfg.Nodes;

with Bwaa.Analysis_Types.Variable_Set; use Bwaa.Analysis_Types.Variable_Set;
with Bwaa.Graph; use type Bwaa.Graph.Vertex_Array_Ptr;
with Bwaa.Util; use Bwaa.Util;
with Bwaa.Options;
use Bwaa;

package body Bwaa.Report is

   procedure Output (Bwaa_Cfg: in Bwaa.Graph.Cfg) is
      Cand_Cur: Variable_Set.Cursor;
      Skip_Reporting: Boolean := False;
   begin

      for I in Bwaa_Cfg.L_Forest.Loops'Range loop
         if Bwaa_Cfg.L_Forest.Loops(I).Terms /= null then
            for J in Bwaa_Cfg.L_Forest.Loops(I).Terms'Range loop

               Skip_Reporting := False;

               if Options.Light_Scan then
                  declare
                     Innocent: Variable_Set.Set := Empty_Set;
                     Innocent_Cur: Variable_Set.Cursor;
                     Cand: Variable;
                  begin
                     Cand_Cur := Variable_Set.First(Bwaa_Cfg.L_Forest.Loops(I).Candidates(J));
                     while Cand_Cur /= Variable_Set.No_Element loop
                        Cand := Variable_Set.Element(Cand_Cur);

                        if not Is_Busy_Var(Cand) then
                           Variable_Set.Insert(Innocent, Cand);

                        elsif Unbounded_String'(Get_Direct_Candidate_Name(Cand)) /= Null_Unbounded_String then

                           Innocent_Cur := Variable_Set.First(Innocent);
                           while Innocent_Cur /= Variable_Set.No_Element loop
                              if Unbounded_String'(Get_Name(Variable_Set.Element(Innocent_Cur))) = Unbounded_String'(Get_Direct_Candidate_Name(Cand)) then
                                 Variable_Set.Delete(Innocent, Innocent_Cur);
                                 exit;
                              end if;
                           end loop;

                        end if;

                        Variable_Set.Next(Cand_Cur);
                     end loop;

                     if not Variable_Set.Is_Empty(Innocent) then
                        Skip_Reporting := True;
                     end if;
                  end;
               end if;

               if not Skip_Reporting then
                  Cand_Cur := Variable_Set.First(Bwaa_Cfg.L_Forest.Loops(I).Candidates(J));
                  while Cand_Cur /= Variable_Set.No_Element loop
                     if Is_Busy_Var(Variable_Set.Element(Cand_Cur)) then
                        Report_Busy_Var(Ast2Cfg.Nodes.Get_Element(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.L_Forest.Loops(I).Header).all),
                                        Ast2Cfg.Nodes.Get_Element(Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.L_Forest.Loops(I).Terms(J)).all),
                                        Variable_Set.Element(Cand_Cur));
                     end if;
                     Variable_Set.Next(Cand_Cur);
                  end loop;
               end if;
            end loop;
         end if;
      end loop;

   end Output;

   procedure Report_Busy_Var (Loop_Element: in Asis.Element;
                              Termination_Element: in Asis.Element;
                              Busy_Var: in Variable) is

      Loop_First_Line_Number: Line_Number := First_Line_Number(Loop_Element);
      Termination_First_Line_Number: Line_Number := First_Line_Number(Termination_Element);
      Enclosing_Element: Asis.Element := Asis.Elements.Enclosing_Element(Loop_Element);
      Report_Line: Unbounded_String := Null_Unbounded_String;
   begin

      if Unbounded_String'(Get_Direct_Candidate_Name(Busy_Var)) /= Null_Unbounded_String then
         Report_Line :=  S2u("[Bwaa] Potential indirect busy waiting with variable " &
                         Get_Name(Busy_Var) & ", direct candidate " & Get_Direct_Candidate_Name(Busy_Var));
      else
         Report_Line :=  S2u("[Bwaa] Potential busy waiting with variable " & Get_Name(Busy_Var));
      end if;

      Report_Line := Report_Line & " in line" &
                     Line_Number_Positive'Image(Termination_First_Line_Number) &
                     " in loop starting at line" &
                     Line_Number_Positive'Image(Loop_First_Line_Number) &
                     " in " & W2s(Asis.Compilation_Units.Unit_Full_Name(Asis.Elements.Enclosing_Compilation_Unit(Loop_Element))) &
                     " in " & W2s(Asis.Compilation_Units.Text_Name(Asis.Elements.Enclosing_Compilation_Unit(Loop_Element)));

      Put_Line(Msg, U2s(Report_Line));

      if Is_Text_Available(Enclosing_Element) then
         New_Line(PrettyMsg);
         declare
            Loop_Print_Start: Integer := Loop_First_Line_Number - (LINE_SPAN / 2);
            Loop_Print_Stop: Line_Number := Loop_Print_Start + LINE_SPAN;
            Termination_Print_Start: Integer := Termination_First_Line_Number - (LINE_SPAN / 2);
            Termination_Print_Stop: Line_Number := Termination_Print_Start + LINE_SPAN;
            Enclosing_First_Line_Number: Line_Number := First_Line_Number(Enclosing_Element);
            Enclosing_Last_Line_Number: Line_Number := Last_Line_Number(Enclosing_Element);
         begin

            if Loop_Print_Start < 0 then
               Loop_Print_Start := 0;
            end if;

            if Termination_Print_Start < 0 then
               Termination_Print_Start := 0;
            end if;

            if Loop_Print_Start < Enclosing_First_Line_Number then
               Loop_Print_Start := Enclosing_First_Line_Number;

               if Loop_Print_Start + LINE_SPAN > Enclosing_Last_Line_Number then
                  Loop_Print_Stop := Enclosing_Last_Line_Number;
               else
                  Loop_Print_Stop := Loop_Print_Start + LINE_SPAN;
               end if;

            elsif Loop_Print_Stop > Enclosing_Last_Line_Number then
               if Enclosing_Last_Line_Number > LINE_SPAN then
                  Loop_Print_Start := Enclosing_Last_Line_Number - LINE_SPAN;
                  Loop_Print_Stop := Loop_Print_Start + LINE_SPAN;
               else
                  Loop_Print_Start := Enclosing_First_Line_Number;
                  Loop_Print_Stop := Enclosing_Last_Line_Number;
               end if;
            end if;

            if Termination_Print_Stop > Enclosing_Last_Line_Number then
               if Enclosing_Last_Line_Number > LINE_SPAN then
                  Termination_Print_Start := Enclosing_Last_Line_Number - LINE_SPAN;
                  Termination_Print_Stop := Termination_Print_Start + LINE_SPAN;
               else
                  Termination_Print_Start := Enclosing_First_Line_Number;
                  Termination_Print_Stop := Enclosing_Last_Line_Number;
               end if;
            end if;

            if Termination_Print_Start <= Loop_Print_Start + LINE_SPAN then
               declare
                  Print_Lines: Line_List := Lines(Enclosing_Element, Loop_Print_Start, Termination_Print_Stop);
               begin
                  for I in Print_Lines'Range loop
                     Put_Line(PrettyMsg, Line_Number'Image(I) & ":" & ASCII.HT & W2s(Line_Image(Print_Lines(I))));
                  end loop;
               end;
            else
               declare
                  Print_Lines: Line_List := Lines(Enclosing_Element, Loop_Print_Start, Loop_Print_Stop);
               begin
                  for I in Print_Lines'Range loop
                     Put_Line(PrettyMsg, Line_Number'Image(I) & ":" & ASCII.HT & W2s(Line_Image(Print_Lines(I))));
                  end loop;
               end;
               Put_Line(PrettyMsg, " ...");
               declare
                  Print_Lines: Line_List := Lines(Enclosing_Element, Termination_Print_Start, Termination_Print_Stop);
               begin
                  for I in Print_Lines'Range loop
                     Put_Line(PrettyMsg, Line_Number'Image(I) & ":" & ASCII.HT & W2s(Line_Image(Print_Lines(I))));
                  end loop;
               end;
            end if;
         end;
            New_Line(PrettyMsg);
      end if;

      Put_Line(PrettyMsg, 70 * "-");

   end Report_Busy_Var;
end Bwaa.Report;
