-- $Id: bwaa-tarjan.adb 136 2009-07-11 15:41:59Z georg $
---------------------------------------------------------------------------
--                                                                       --
--                   BWAA - Busy Wait Analyser for Ada                   --
--      BWAA is a program that finds busy waiting in Ada programs.       --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--                 Copyright (c) 2009 Georg Kienesberger                 --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of BWAA.                                            --
--                                                                       --
-- BWAA is free software; you can redistribute it and/or modify          --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- BWAA is distributed in the hope that it will be useful,               --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ast2Cfg; use type Ast2Cfg.Long_Long_Natural;

with Bwaa.Graph; use Bwaa.Graph;

use type Vertex_Set.Cursor;
use type Vertex_Map.Cursor;

package body Bwaa.Tarjan is

   procedure Collapse (Cfg: in out Graph.Cfg; Loop_Body: in Vertex_Set.Set; Loop_Header: in Ast2Cfg.Long_Long_Natural) is
      LB_Cur, Nested_Cur: Vertex_Set.Cursor;
      Nested_Header, Nested_Num: Ast2Cfg.Long_Long_Natural := 0;
      Nested: Graph.L00p;
   begin
      -- increase Loop_Count which also serves as an id for each loop
      Loop_Count := Loop_Count + 1;
      if Loop_Count > Cfg.L_Forest.Loop_Count then
         raise Bwaa_Exception with "Tarjan: Encountered more loops than expected!";
      end if;

      -- add this new loop to the forest
      Cfg.L_Forest.Loops(Loop_Count).Header := Loop_Header;
      Cfg.L_Forest.Loops(Loop_Count).B0dy := Loop_Body;
      -- remember which header belongs to this loop number
      Vertex_Map.Insert(Header_To_Loop, Loop_Header, Loop_Count);

      -- for every node in the loop body...
      LB_Cur := Vertex_Set.First(Loop_Body);
      while LB_Cur /= Vertex_Set.No_Element loop
         -- ...collapse it on the header; we may use link here (instead of union),
         -- since we only get canonical elements from Findloop
         Uf.Link(Uf_Sets, Vertex_Set.Element(LB_Cur), Loop_Header);

         -- in case the current node is a loop header too...
         Nested_Header := Vertex_Set.Element(LB_Cur);
         if Vertex_Map.Contains(Header_To_Loop, Nested_Header) then
            -- ...record that nesting relation...
            Nested_Num := Vertex_Map.Element(Header_To_Loop, Nested_Header);
            Cfg.L_Forest.Preds(Nested_Num) := Loop_Count;

            -- ...and add the nodes from the nested loop
            Nested := Cfg.L_Forest.Loops(Nested_Num);
            Nested_Cur := Vertex_Set.First(Nested.B0dy);
            while Nested_Cur /= Vertex_Set.No_Element loop
               Vertex_Set.Insert(Cfg.L_Forest.Loops(Loop_Count).B0dy, Vertex_Set.Element(Nested_Cur));
               Vertex_Set.Next(Nested_Cur);
            end loop;
         end if;

         Vertex_Set.Next(LB_Cur);
      end loop;
   end Collapse;

   procedure Findloop (Cfg: in out Graph.Cfg; Potential_Header: in Ast2Cfg.Long_Long_Natural) is
      Worklist, Loop_Body: Vertex_Set.Set;
      Canonical, Work: Ast2Cfg.Long_Long_Natural;
      Work_Cur: Vertex_Set.Cursor;
      BE_Cur: Vertex_Map.Cursor;
   begin
      -- is the potential header really a loop header (i.e. a target of a backedge)?
      if Vertex_Set.Contains(Cfg.Headers, Potential_Header) then

         -- find the canonical elements of the nodes where the backedges to the header originate and add them to the worklist
         BE_Cur := Vertex_Map.First(Cfg.Backedges);
         while BE_Cur /= Vertex_Map.No_Element loop
            if Vertex_Map.Element(BE_Cur) = Potential_Header then
               Uf.Find(Uf_Sets, Vertex_Map.Key(BE_Cur), Canonical);
               Vertex_Set.Insert(Worklist, Canonical);
            end if;
            Vertex_Map.Next(BE_Cur);
         end loop;

         -- pick an element from the worklist until it is empty
         Work_Cur := Vertex_Set.First(Worklist);
         while Work_Cur /= Vertex_Set.No_Element loop
            Work := Vertex_Set.Element(Work_Cur);
            Vertex_Set.Delete(Worklist, Work_Cur);

            -- add the current element to the loop body
            Vertex_Set.Insert(Loop_Body, Work);

            -- for each predecessor that is not origin of a backedge...
            if Cfg.Preds(Work) /= null then
               for I in Cfg.Preds(Work)'Range loop
                  if not Vertex_Set.Contains(Cfg.Headers, Work) or else
                    (not Vertex_Map.Contains(Cfg.Backedges, Cfg.Preds(Work)(I)) or else
                     Vertex_Map.Element(Cfg.Backedges, Cfg.Preds(Work)(I)) /= Work) then

                     -- ...find the canonical element, and add it to the worklist in case
                     -- it is not already in the loop body, and is different from the header
                     Uf.Find(Uf_Sets, Cfg.Preds(Work)(I), Canonical);
                     if not Vertex_Set.Contains(Loop_Body, Canonical) and
                            Canonical /= Potential_Header then
                        -- we do not have to check if it is already on the worklist, since
                        -- include does no harm in case it is already there
                        Vertex_Set.Include(Worklist, Canonical);
                     end if;
                  end if;
               end loop;
            end if;
            Work_Cur := Vertex_Set.First(Worklist);
         end loop;
         -- in case we found a loop call Collapse
         if not Vertex_Set.Is_Empty(Loop_Body) then
            Collapse(Cfg, Loop_Body, Potential_Header);
         end if;
      end if;
   end Findloop;

   -- construct the loop forest (Cfg.L_Forest) using a slightly modified version of Tarjans algorithm
   procedure Build_Loop_Forest (Cfg: in out Graph.Cfg) is
   begin

      -- place every node in a set by itself
      Uf.Init_Sets(Uf_Sets);

      -- construct an empty loop forest; the number of the loop headers
      -- (targets of backedges) is the number of loops in the cfg
      Cfg.L_Forest := new Graph.Loop_Forest(Ast2Cfg.Long_Long_Natural(Vertex_Set.Length(Cfg.Headers)));
      Graph.Init(Cfg.L_Forest.all);

      -- in reverse dfs order call Findloop for every node
      for I in reverse 1 .. Cfg.Node_Count loop
         Findloop(Cfg, I);
      end loop;

      -- clean up
      Uf.Init_Sets(Uf_Sets);
      Loop_Count := 0;
      Vertex_Map.Clear(Header_To_Loop);
   end Build_Loop_Forest;

end Bwaa.Tarjan;
