-- $Id: bwaa-traversal.adb 186 2009-07-28 06:38:38Z georg $
---------------------------------------------------------------------------
--                                                                       --
--                   BWAA - Busy Wait Analyser for Ada                   --
--      BWAA is a program that finds busy waiting in Ada programs.       --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--                 Copyright (c) 2009 Georg Kienesberger                 --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of BWAA.                                            --
--                                                                       --
-- BWAA is free software; you can redistribute it and/or modify          --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- BWAA is distributed in the hope that it will be useful,               --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ast2Cfg.Pkgs;
with Ast2Cfg.Cfgs;
with Ast2Cfg.Nodes; use Ast2Cfg.Nodes;
with Ast2Cfg.Flow_Types; use Ast2Cfg.Flow_Types;
with Ast2Cfg.Id;
with Ast2cfg.Declaration_Id;
with Ast2Cfg.Utils;
use Ast2Cfg;
use type Ast2Cfg.Long_Long_Natural;

with Bwaa.Output; use Bwaa.Output;

package body Bwaa.Traversal is

   -- Fill Bwaa_Cfg.Dfs_Sorted and Bwaa_Cfg.Id_Dfs, find backedges
   procedure Dfs (Ast2Cfg_Cfg: in Ast2Cfg.Cfgs.CFG_Object; Bwaa_Cfg: out Bwaa.Graph.Cfg) is

      procedure Recursion (Cur: in Node_Class_Ptr; Last: in Ast2Cfg.Long_Long_Natural) is
         Succs: Flow_Class_Ptr_List.Object;
         Succ: Node_Class_Ptr := null;
         Cursor: Id_To_Dfs_Num.Cursor;
         Cur_Num: Ast2Cfg.Long_Long_Natural;
      begin

         -- check the node mark
         if Cur = null or else Get_Mark(Cur.all) = 2 then
            return;
         elsif Get_Mark(Cur.all) = 1 then
            -- we found a backedge
            Cursor := Id_To_Dfs_Num.Find(Bwaa_Cfg.Id_Dfs, Get_Id(Cur.all));
            Bwaa.Graph.Vertex_Map.Insert(Bwaa_Cfg.Backedges, Last, Id_To_Dfs_Num.Element(Cursor));
            Bwaa.Graph.Vertex_Set.Include(Bwaa_Cfg.Headers, Id_To_Dfs_Num.Element(Cursor));
            if Last = 0 then
               raise Bwaa_Exception with "Traversal: Added backedge starting at node number 0!";
            end if;
            return;
         elsif Get_Mark(Cur.all) /= 0 and Get_Mark(Cur.all) /= 2 then
            raise Bwaa_Exception with "Traversal: Node Marks not reset correctly!";
         end if;

         -- check if there is enough space for the current node
         Node_Count := Node_Count + 1;
         Cur_Num := Node_Count;
         if Node_Count > Bwaa_Cfg.Node_Count then
            raise Bwaa_Exception with "Traversal: Actual node count (" &
                                       Ast2Cfg.Long_Long_Natural'Image(Node_Count) &
                                       ") exceeds reserved array (" &
                                       Ast2Cfg.Long_Long_Natural'Image(Bwaa_Cfg.Node_Count) &
                                       ")!";
         end if;

         -- set the mark of the current node
         Set_Mark(Cur.all, 1);

         Put_Line(Max, "[Bwaa.Dfs] at Node " & Id.To_String(Get_Id(Cur.all)));

         -- add node pointer to sorted array
         Bwaa_Cfg.Dfs_Sorted(Bwaa_Cfg.Dfs_Sorted'First + Node_Count - 1) := Cur;

         -- map the node id to the dfs number
         Id_To_Dfs_Num.Insert(Bwaa_Cfg.Id_Dfs, Get_Id(Cur.all), Node_Count);

         -- recurse for all successors
         Succs := Get_Succs(Cur.all);
         Flow_Class_Ptr_List.Reset(Succs);
         while Flow_Class_Ptr_List.Has_Next(Succs) loop
            Flow_Class_Ptr_List.Get_Next(Succs, Flow_Class_Ptr(Succ));
            Recursion(Succ, Cur_Num);
         end loop;

         -- recursion for this node finished; mark it with 2
         Set_Mark(Cur.all, 2);
      end Recursion;

      Root: Node_Class_Ptr := Cfgs.Get_Head(Ast2Cfg_Cfg);
   begin
      Reset_Mark(Root);
      Recursion(Root, 0);

      if Node_Count /= Bwaa_Cfg.Node_Count then
         raise Bwaa_Exception with "Traversal: Expected " & Ast2Cfg.Long_Long_Natural'Image(Bwaa_Cfg.Node_Count) & " Nodes, not " & Ast2Cfg.Long_Long_Natural'Image(Node_Count);
      end if;
      Node_Count := 0;
   end Dfs;

   -- Find all specifications and fill a hashed map with references to them
   procedure Collect_Specs (Flow: in Ast2Cfg.Flow_Types.Flow_Class_Ptr;
                            Specs: in out Specification_Map.Map;
                            Path: in Unbounded_String) is

      Name: Unbounded_String := Null_Unbounded_String;
      Succs: Ast2Cfg.Flow_Types.Flow_Class_Ptr_List.Object := Get_Succs(Flow.all);
      Succ: Ast2Cfg.Flow_Types.Flow_Class_Ptr := null;
      use Flow_Class_Ptr_List;
   begin

      if not (Flow.all in Ast2Cfg.Pkgs.Def_Object'Class) then

         if Path /= Null_Unbounded_String then
            Name := Path & ".";
         end if;
         Name := Name & Unbounded_String'(Get_Name(Flow.all));

         if Flow.all in Ast2Cfg.Pkgs.Spec_Object'Class then
            Put_Line(Debug, "[Bwaa.Traversal] Inserting spec " & U2s(Name) & " into specification map.");

            declare
            begin
               Specification_Map.Insert(Specs, Name, Flow);
            exception
               when Constraint_Error => Put_Line(Error, "[Bwaa.Traversal] ERROR: Duplicate specification " & U2s(Name) & " encountered.");
            end;
         end if;

      end if;

      Reset(Succs);
      while Has_Next(Succs) loop
         Get_Next(Succs, Succ);
         Collect_Specs(Succ, Specs, Name);
      end loop;

   end Collect_Specs;

   -- collect the variables, packages and renamings that are visible in each flow
   procedure Build_Visibility_Info (Flow: in Ast2Cfg.Flow_Types.Flow_Class_Ptr;
                                    Visi_Info: in out Visibility_Info.Info;
                                    Passed_On: in Visible_Set.Set;
                                    Pkg_Passed_On: in Visible_Set.Set;
                                    Spec_Map: in Bwaa.Util.Specification_Map.Map) is

      use Ast2Cfg.Utils;
      use Flow_Class_Ptr_List;
      use Visible_Set;

      Succs: Ast2Cfg.Flow_Types.Flow_Class_Ptr_List.Object := Get_Succs(Flow.all);
      Succ: Ast2Cfg.Flow_Types.Flow_Class_Ptr := null;
      Decls: Declaration_Ptr_List.Object := Get_Decls(Flow.all);
      Rens: Alias_Class_Ptr_List.Object := Get_Rens(Flow.all);
      Cur_Decl: Declaration_Ptr := null;
      Cur_Ren: Alias_Class_Ptr := null;
      Var_Name: Unbounded_String := Null_Unbounded_String;
      Local: Visible_Set.Set;
      Pkg_Local: Visible_Set.Set;
      Visible: Visible_Set.Set;
      Pkg_Visible: Visible_Set.Set;
      Inherited: Visible_Set.Set := Passed_On;
      Pkg_Inherited: Visible_Set.Set := Pkg_Passed_On;
   begin

      -- is this flow already visited?
      if (Flow.all in Ast2Cfg.Pkgs.Spec_Object'Class) or
         (Flow.all in Ast2Cfg.CFGs.Spec_Object'Class) then

         if Visibility_Map.Contains(Visi_Info.Specs, Get_Full_Name(Flow.all)) or else
            Visibility_Map.Contains(Visi_Info.Pkg_Specs, Get_Full_Name(Flow.all)) then

            Put_Line(Debug, "[Bwaa.Traversal] Build_Visibility_Map: Already handled " & Bwaa.Util.U2s(Get_Full_Name(Flow.all)) & " (spec)");
            return;
         end if;

      elsif (Flow.all in Ast2Cfg.Pkgs.Body_Object'Class) or
            (Flow.all in Ast2Cfg.CFGs.Body_Object'Class) then

         if Visibility_Map.Contains(Visi_Info.Bodies, Get_Full_Name(Flow.all)) or else
            Visibility_Map.Contains(Visi_Info.Pkg_Bodies, Get_Full_Name(Flow.all)) then

            Put_Line(Debug, "[Bwaa.Traversal] Build_Visibility_Map: Already handled " & Bwaa.Util.U2s(Get_Full_Name(Flow.all)) & " (body)");
            return;
         end if;

      else
         raise Visibility_Exception with "[Bwaa.Traversal] ERROR: Build_Visibility_Map: Flow is neither spec nor body.";
      end if;

      -- get the variable/package declarations/renamings of the spec of the current flow object
      if not (Flow.all in Ast2Cfg.Pkgs.Spec_Object'Class) and
         not (Flow.all in Ast2Cfg.CFGs.Spec_Object'Class) then

         if Visibility_Map.Contains(Visi_Info.Specs, Get_Full_Name(Flow.all)) then
            Put_Line(Debug, "[Bwaa.Traversal] Build_Visibility_Map: Merging visibility information from " & Bwaa.Util.U2s(Get_Full_Name(Flow.all)) & " (spec) into body.");
            declare
               Spec_Vars: Visible_Set.Set := Visibility_Map.Element(Visi_Info.Specs, Get_Full_Name(Flow.all));
               Pkg_Spec_Vars: Visible_Set.Set := Visibility_Map.Element(Visi_Info.Pkg_Specs, Get_Full_Name(Flow.all));
            begin
               Inherited := Spec_Vars or (Inherited - Spec_Vars);
               Pkg_Inherited := Pkg_Spec_Vars or (Pkg_Inherited - Pkg_Spec_Vars);
            end;
         else
            if Specification_Map.Contains(Spec_Map, Get_Full_Name(Flow.all)) then
               Put_Line(Debug, "[Bwaa.Traversal] Build_Visibility_Map: Need to jump to " & Bwaa.Util.U2s(Get_Full_Name(Flow.all)) & " (spec).");
               declare
                  Spec: Flow_Class_Ptr := Specification_Map.Element(Spec_Map, Get_Full_Name(Flow.all));
                  Preds: Flow_Class_Ptr_List.Object := Get_Preds(Spec.all);
                  Root: Flow_Class_Ptr := Spec;
               begin

                  -- search root in Pkg/CFG tree
                  SEARCH:
                  loop
                     if Count(Preds) > 1 then
                        raise Visibility_Exception with "[Bwaa.Traversal] ERROR: Build_Visibility_Map: More than one predecessor found.";
                     elsif Count(Preds) = 1 then
                        Root := Get_Head(Preds);
                        Preds := Get_Preds(Root.all);
                     else
                          exit SEARCH;
                     end if;
                  end loop SEARCH;

                  -- start new recursion at root
                  Build_Visibility_Info(Root, Visi_Info, Visible_Set.Empty_Set, Visible_Set.Empty_Set, Spec_Map);

                  -- now there should be the visible set of the spec in the visibility map
                  if (not Visibility_Map.Contains(Visi_Info.Specs, Get_Full_Name(Flow.all))) or else
                     (not Visibility_Map.Contains(Visi_Info.Pkg_Specs, Get_Full_Name(Flow.all))) then

                     raise Visibility_Exception with "[Bwaa.Traversal] ERROR: Unable to get visible set of spec for " & Bwaa.Util.U2s(Get_Full_Name(Flow.all));
                  end if;

                  Put_Line(Debug, "[Bwaa.Traversal] Build_Visibility_Map: Merging visibility information from " & Bwaa.Util.U2s(Get_Full_Name(Flow.all)) & " (spec) into body.");
                  declare
                     Spec_Vars: Visible_Set.Set := Visibility_Map.Element(Visi_Info.Specs, Get_Full_Name(Flow.all));
                     Pkg_Spec_Vars: Visible_Set.Set := Visibility_Map.Element(Visi_Info.Pkg_Specs, Get_Full_Name(Flow.all));
                  begin
                     Inherited := Spec_Vars or (Inherited - Spec_Vars);
                     Pkg_Inherited := Pkg_Spec_Vars or (Pkg_Inherited - Pkg_Spec_Vars);
                  end;
               end;
            end if;
         end if;

      end if;

      -- get the parameter declarations of the current flow object in case it is a subprogram
      if Flow.all in Ast2Cfg.CFGs.Body_Object'Class then
         declare
            Params: Declaration_Ptr_List.Object := Ast2Cfg.CFGs.Get_Params(Ast2Cfg.CFGs.CFG_Object(Flow.all));
            Cur_Param: Declaration_Ptr := null;
            Param_Name: Unbounded_String := Null_Unbounded_String;
         begin
            Declaration_Ptr_List.Reset(Params);
            while Declaration_Ptr_List.Has_Next(Params) loop
               Declaration_Ptr_List.Get_Next(Params, Cur_Param);

               UStr_List.Reset(Cur_Param.Vars);
               while Ustr_List.Has_Next(Cur_Param.Vars) loop
                  Ustr_List.Get_Next(Cur_Param.Vars, Param_Name);
                  declare
                     New_Decl: Bwaa.Visibility.Declaration;
                  begin
                     Init(New_Decl, Param_Name, Get_Full_Name(Flow.all), Cur_Param.Decl_Id);
                     Visible_Set.Insert(Local, New_Decl);
                  exception
                     when Constraint_Error => Put_Line(Debug, "[Bwaa.Traversal] Duplicate declaration " & Bwaa.Util.U2s(Var_Name) & " in " & Bwaa.Util.U2s(Get_Full_Name(Flow.all)) & " encountered.");
                  end;
               end loop;
            end loop;
         end;
      end if;

      -- get the variable declarations of the current flow object
      Declaration_Ptr_List.Reset(Decls);
      while Declaration_Ptr_List.Has_Next(Decls) loop
         Declaration_Ptr_List.Get_Next(Decls, Cur_Decl);

         UStr_List.Reset(Cur_Decl.Vars);
         while Ustr_List.Has_Next(Cur_Decl.Vars) loop
            Ustr_List.Get_Next(Cur_Decl.Vars, Var_Name);
            declare
               New_Decl: Bwaa.Visibility.Declaration;
            begin
               Init(New_Decl, Var_Name, Get_Full_Name(Flow.all), Cur_Decl.Decl_Id);
               Visible_Set.Insert(Local, New_Decl);
            exception
               when Constraint_Error => Put_Line(Debug, "[Bwaa.Traversal] Duplicate declaration " & Bwaa.Util.U2s(Var_Name) & " in " & Bwaa.Util.U2s(Get_Full_Name(Flow.all)) & " encountered. (forward declaration?)");
            end;
         end loop;
      end loop;

      -- add the current flow in case it is a package
      if Flow.all in Ast2Cfg.Pkgs.Pkg_Object'Class then
         declare
            New_Decl: Bwaa.Visibility.Declaration;
         begin
            Init(New_Decl, Get_Name(Flow.all), Get_Prefix(Flow.all), Ast2Cfg.Pkgs.Get_Declaration_Id(Ast2Cfg.Pkgs.Pkg_Object(Flow.all)));
            Visible_Set.Include(Pkg_Local, New_Decl);
         end;
      end if;

      -- get the declared packages of the current flow object
      Reset(Succs);
      while Has_Next(Succs) loop
         Get_Next(Succs, Succ);
         if Succ.all in Ast2Cfg.Pkgs.Pkg_Object'Class then
            declare
               New_Decl: Bwaa.Visibility.Declaration;
            begin
               Init(New_Decl, Get_Name(Succ.all), Get_Full_Name(Flow.all), Ast2Cfg.Pkgs.Get_Declaration_Id(Ast2Cfg.Pkgs.Pkg_Object(Succ.all)));
               Visible_Set.Insert(Pkg_Local, New_Decl);
            exception
               when Constraint_Error => Put_Line(Debug, "[Bwaa.Traversal] Duplicate declaration " & Bwaa.Util.U2s(Var_Name) & " in " & Bwaa.Util.U2s(Get_Full_Name(Flow.all)) & " encountered. (forward declaration?)");
            end;
         end if;
      end loop;

      -- get the renamings of the current flow object
      Alias_Class_Ptr_List.Reset(Rens);
      while Alias_Class_Ptr_List.Has_Next(Rens) loop
         Alias_Class_Ptr_List.Get_Next(Rens, Cur_Ren);

         if Cur_Ren.all in Obj_Alias'Class then
            declare
               New_Decl: Bwaa.Visibility.Declaration;
            begin
               Init(New_Decl, Cur_Ren.New_Name, Get_Full_Name(Flow.all), Cur_Ren.Decl_Id);
               Visible_Set.Insert(Local, New_Decl);
            exception
               when Constraint_Error => Put_Line(Error, "[Bwaa.Traversal] Duplicate declaration " & Bwaa.Util.U2s(Cur_Ren.New_Name) & " in " & Bwaa.Util.U2s(Get_Full_Name(Flow.all)) & " encountered. (forward declaration?)");
            end;

         elsif Cur_Ren.all in Pkg_Alias'Class then
            declare
               New_Decl: Bwaa.Visibility.Declaration;
            begin
               Init(New_Decl, Cur_Ren.New_Name, Get_Full_Name(Flow.all), Cur_Ren.Decl_Id);
               Visible_Set.Insert(Pkg_Local, New_Decl);
            exception
               when Constraint_Error => Put_Line(Error, "[Bwaa.Traversal] Duplicate declaration " & Bwaa.Util.U2s(Cur_Ren.New_Name) & " in " & Bwaa.Util.U2s(Get_Full_Name(Flow.all)) & " encountered. (forward declaration?)");
            end;
         end if;

      end loop;

      -- save the still visible variables/packages/renamings from above and the local ones into one set
      Visible := Local or (Inherited - Local);
      Pkg_Visible := Pkg_Local or (Pkg_Inherited - Pkg_Local);

      -- save the visible sets for this flow
      declare
      begin
         if (Flow.all in Ast2Cfg.Pkgs.Spec_Object'Class) or
            (Flow.all in Ast2Cfg.CFGs.Spec_Object'Class) then

            Put_Line(Debug, "[Bwaa.Traversal] Build_Visibility_Map: Inserting visible sets for " & Bwaa.Util.U2s(Get_Full_Name(Flow.all)) & " (spec)");
            Visibility_Map.Insert(Visi_Info.Specs, Get_Full_Name(Flow.all), Visible);
            Visibility_Map.Insert(Visi_Info.Pkg_Specs, Get_Full_Name(Flow.all), Pkg_Visible);
         elsif (Flow.all in Ast2Cfg.Pkgs.Body_Object'Class) or
               (Flow.all in Ast2Cfg.CFGs.Body_Object'Class) then

            Put_Line(Debug, "[Bwaa.Traversal] Build_Visibility_Map: Inserting visible sets for " & Bwaa.Util.U2s(Get_Full_Name(Flow.all)) & " (body)");
            Visibility_Map.Insert(Visi_Info.Bodies, Get_Full_Name(Flow.all), Visible);
            Visibility_Map.Insert(Visi_Info.Pkg_Bodies, Get_Full_Name(Flow.all), Pkg_Visible);
         else
            raise Visibility_Exception with "[Bwaa.Traversal] ERROR: Building visibility map: Flow is neither spec nor body.";
         end if;

      exception
         when Constraint_Error => Put_Line(Error, "[Bwaa.Traversal] ERROR: At least one visible set for " & Bwaa.Util.U2s(Get_Full_Name(Flow.all)) & " already saved.");
      end;

      -- recursion
      Reset(Succs);
      while Has_Next(Succs) loop
         Get_Next(Succs, Succ);

         declare
            use Visible_Set;
            Inherit: Visible_Set.Set := Visible_Set.Empty_Set;
            Pkg_Inherit: Visible_Set.Set := Visible_Set.Empty_Set;
            Cur_Local: Cursor := First(Local);
            Cur_Pkg_Local: Cursor := First(Pkg_Local);
            Flow_Decl_Id: Ast2Cfg.Declaration_Id.Object := Ast2Cfg.Declaration_Id.Null_Declaration_Id;
         begin

            if Succ.all in Ast2Cfg.Pkgs.Pkg_Object'Class then
               Flow_Decl_Id := Ast2Cfg.Pkgs.Get_Declaration_Id(Ast2Cfg.Pkgs.Pkg_Object(Succ.all));
            elsif Succ.all in Ast2Cfg.CFGs.CFG_Object'Class then
               Flow_Decl_Id := Ast2Cfg.CFGs.Get_Declaration_Id(Ast2Cfg.CFGs.CFG_Object(Succ.all));
            else
               raise Visibility_Exception with "[Bwaa.Traversal] ERROR: Unexpected flow while building visibility map.";
            end if;

            -- determine which local variables are inherited
            while Cur_Local /= No_Element loop

               if Element(Cur_Local) < Flow_Decl_Id then
                  Insert(Inherit, Element(Cur_Local));
               end if;

               Next(Cur_Local);
            end loop;

            -- important: only those variables can overwrite inherited ones that are declared before the succ
            Inherit := Inherit or (Inherited - Inherit);

            while Cur_Pkg_Local /= No_Element loop

               if Element(Cur_Pkg_Local) < Flow_Decl_Id then
                  Insert(Pkg_Inherit, Element(Cur_Pkg_Local));
               end if;

               Next(Cur_Pkg_Local);
            end loop;

            -- important: only those packages can overwrite inherited ones that are declared before the succ
            Pkg_Inherit := Pkg_Inherit or (Pkg_Inherited - Pkg_Inherit);

            Build_Visibility_Info(Succ, Visi_Info, Inherit, Pkg_Inherit, Spec_Map);
         end;
      end loop;

   end Build_Visibility_Info;

end Bwaa.Traversal;
