-- $Id: bwaa-util.adb 186 2009-07-28 06:38:38Z georg $
---------------------------------------------------------------------------
--                                                                       --
--                   BWAA - Busy Wait Analyser for Ada                   --
--      BWAA is a program that finds busy waiting in Ada programs.       --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--                 Copyright (c) 2009 Georg Kienesberger                 --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of BWAA.                                            --
--                                                                       --
-- BWAA is free software; you can redistribute it and/or modify          --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- BWAA is distributed in the hope that it will be useful,               --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Strings.Hash;
with Ada.Characters.Handling;

package body Bwaa.Util is

   -- "... if this doesn't work well then we can blame the vendor." John Barnes, Programming in Ada 2005
   function Hash (X: in Ast2Cfg.Long_Long_Natural) return Ada.Containers.Hash_Type is
   begin
      return Ada.Strings.Hash(Ast2Cfg.Long_Long_Natural'Image(X));
   end Hash;

   -- Convert a String to an Unbounded_String
   function S2u(Str: in String) return Unbounded_String is
   begin
      return Ada.Strings.Unbounded.To_Unbounded_String(Str);
   end S2u;

   -- Convert an Unbounded_String to a String
   function U2s(U_Str: in Unbounded_String) return String is
   begin
      return Ada.Strings.Unbounded.To_String(U_Str);
   end U2s;

   -- Convert a Wide_String to a String
   function W2s(W_Str: in Wide_String) return String is
   begin
      return Ada.Characters.Handling.To_String(W_Str);
   end W2s;

   -- Split an Unbounded_String into multiple Unbounded_Strings based on a separator
   function Split(U: in Unbounded_String; Separator: in String := ".") return Unbounded_String_Array is
      Partition_Number: Natural := Count(U, ".") + 1;
      Partition: Unbounded_String_Array(1 .. Partition_Number);
      Cur_Low: Natural := 0;
      Cur_High: Natural := 0;
   begin

      if Partition_Number < 2 then
         Partition(1) := U;
         return Partition;
      end if;

      for I in Partition'Range loop
         Cur_Low := Cur_High;
         Cur_High := Index(U, Separator, Cur_High + 1);
         if Cur_High = 0 then
            if I /= Partition'Last then
               raise Bwaa_Exception with "Unable to find separator.";
            end if;
            Cur_High := Length(U) + 1;
         end if;

         Partition(I):= Unbounded_Slice(U, Cur_Low + 1, Cur_High - 1);
      end loop;

      return Partition;

   exception
      when others => raise Bwaa_Exception with "Error while splitting Unbounded_String.";
   end Split;

   -- Merge multiple Unbounded_Strings into a single Unbounded_String using a separator
   function Merge(U: in Unbounded_String_Array;
                  Stop: in Positive;
                  Start: in Positive := 1;
                  Separator: in String := ".") return Unbounded_String is

      Retval: Unbounded_String := Null_Unbounded_String;
      I: Positive := Start;
   begin

      if Stop > U'Last or
         Start > U'Last then

         declare
            Error_Msg: Unbounded_String := S2u("Error while merging Unbounded_Strings: Start/Stop too large: ");
         begin
            for J in U'Range loop
               if J /= U'First then
                  Error_Msg := Error_Msg & Separator;
               end if;
               Error_Msg := Error_Msg & U(J);
            end loop;

            raise Bwaa_Exception with U2s(Error_Msg) &
                                      ", Range:" & Positive'Image(U'First) &
                                      " .." & Positive'Image(U'Last) &
                                      ", Start:" & Positive'Image(Start) &
                                      ", Stop:" & Positive'Image(Stop);
         end;
      end if;

      while I <= Stop loop
         if I /= Start then
            Retval := Retval & Separator;
         end if;
         Retval := Retval & U(I);
         I := I + 1;
      end loop;

      return Retval;
   end Merge;

   -- Merge multiple Unbounded_Strings into a single Unbounded_String using a separator
   function Merge(U: in Unbounded_String_Array) return Unbounded_String is
   begin
      return Merge(U, U'Last);
   end Merge;

end Bwaa.Util;
